import os, signal, json, itertools, traceback, sys
from subprocess import Popen, PIPE, TimeoutExpired
import platform
import logging
import re
from . import opcodes
from . import parse_int_or_hex,decode_hex,remove_0x_head

logger = logging.getLogger()

FNULL = open(os.devnull, 'w')

valid_opcodes = opcodes.reverse_opcodes.keys()

# The 'stateRoot' output is missing from some clients (parity). 
# Enable this once they implement it
INCLUDE_STATEROOT=True

strip_0x = remove_0x_head
bstrToInt = lambda b_str: int(b_str.replace("b", "").replace("'", ""))
bstrToHex = lambda b_str: '0x{0:01x}'.format(bstrToInt(b_str))

def add_0x(str):
    if str in [None, "0x", ""]:
        return ""
    if str[:2] == "0x":
        return str
    return "0x" + str

def toHexQuantities(vals):
    """ Formats a list of values into a list of hex-encoded values """
    return ['0x{0:01x}'.format(parse_int_or_hex(val)) for val in vals]


def toText(op):
    if len(op.keys()) == 0:
        return "END"
    if 'pc' in op.keys():
        op_key = op['op']
        if op_key in opcodes.opcodes.keys():
            opname = opcodes.opcodes[op_key][0]
        else:
            opname = "UNKNOWN"
        op['opname'] = opname
        return "pc {pc:>5} op {opname:>10}({op:>3}) gas {gas:>8} depth {depth:>2} stack {stack}".format(**op)
    elif 'stateRoot' in op.keys():
        return "stateRoot {}".format(op['stateRoot'])
    elif 'time' in op.keys():# Final one

        if 'output' not in op.keys():
           op['output'] = ""

        op['output'] = strip_0x(op['output'])
        fmt = "output {output} gasUsed {gasUsed}"
        if 'error' in op.keys():
            e = op['error']
            if e.lower().find("out of gas") > -1:   
                e = "OOG"
            fmt = fmt + " err: OOG"
        return fmt.format(**op)
    return "N/A"
""" Not used (??)
def getIntrinsicGas(data):
    import ethereum.transactions as transactions
    tx = transactions.Transaction(
        nonce=0,
        gasprice=0,
        startgas=0,
        to=b"",
        value=0,
        data=decode_hex(remove_0x_head(data)))
    
    return tx.intrinsic_gas_used
"""
def compare_traces(clients_canon_traces, names):

    """ Compare 'canonical' traces from the clients"""

    full_output = []
    log = lambda x: full_output.append(x)


    canon_traces = list(itertools.zip_longest(*clients_canon_traces))

    num_clients = len(names)
    equivalent = True
    for step in canon_traces:
        wrong_clients = []
        step_equiv = True
        for i in range(1, num_clients):
            if step[i] != step[0]:
                step_equiv = False
                wrong_clients.append(i)

        if step_equiv == True:
            log('[*] {:>8} {}'.format("", step[0]))
        else:
            equivalent = False
            for i in range(0, num_clients):
                if i in wrong_clients or len(wrong_clients) == num_clients-1:
                    log('[!!] {:>7} {}'.format(names[i], step[i]))
                else:
                    log('[*] {:>8} {}'.format(names[i], step[i]))

    return (equivalent, full_output)


def startProc(cmd):
    # passing a list to Popen doesn't work. Can't read stdout from docker container when shell=False
    #pyeth_process = subprocess.Popen(pyeth_docker_cmd, shell=False, stdout=subprocess.PIPE, close_fds=True)

    # need to pass a string to Popen and shell=True to get stdout from docker container
    print(" ".join(cmd))
    return Popen(" ".join(cmd), stdout=PIPE,shell=True, stderr=PIPE, preexec_fn=os.setsid)


def finishProc(process, extraTime=False, output="stdout", timeout = 30):

    if extraTime:
        timeout = 45
    try:
        (stdoutdata, stderrdata) = process.communicate(timeout=timeout)
    except TimeoutExpired:
        logger.info("TIMEOUT ERROR!")
        os.killpg(process.pid, signal.SIGINT) # send signal to the process group
        (stdoutdata, stderrdata) = process.communicate()

    if output == 'stdout':
        return stdoutdata.decode().strip().split("\n")
    return stderrdata.decode().strip().split("\n")

class VM(object):

    def __init__(self,executable="evmbin", docker = False):
        self.executable = executable
        self.docker = docker
        self.genesis_format = "parity"
        self.lastCommand = ""

    def _run(self,cmd):
        self.lastCommand = " ".join(cmd)
        return finishProc(startProc(cmd))

    def _start(self, cmd):
        self.lastCommand = " ".join(cmd)
        return startProc(cmd)

class JsVM(VM):
    @staticmethod
    def canonicalized(output):
        steps = []
        for index, line in enumerate(output):
            if line and line.startswith('# {'):
                result = json.loads(line.strip('# '))
                steps.append(result)
        return steps


class HeraVM(VM):
    @staticmethod
    def canonicalized(output):
        from . import opcodes
        valid_opcodes = opcodes.reverse_opcodes.keys()

        steps = []
        for x in output:
            try:
                if len(x) > 0  and x[0] == "{":
                    step = json.loads(x)
                    if 'stateRoot' in step.keys() and INCLUDE_STATEROOT:
                      steps.append(step)
                    else:
                      step['gas'] = hex(step['gas'])
                      step['stack'] = step['stack'][::-1]
                      for i in range(0, len(step['stack'])):
                          step['stack'][i] = re.sub(r'0x0+([0-9a-f]+)$', '0x\g<1>', step['stack'][i])

                      steps.append(step)

            except Exception as e:
                logger.info('Exception parsing Hera json:')
                logger.info(e)
                logger.info('problematic line:')
                logger.info(x[:500])

        return steps


class CppVM(VM):

    @staticmethod
    def canonicalized(output):
        from . import opcodes
        valid_opcodes = opcodes.reverse_opcodes.keys()

        steps = []
        for x in output:
            try:
                if x[0:2] == "[{":
                        steps = json.loads(x)

                if x[0:2] == "{\"":
                    # A bug in testeth
                    if x[-1] == '.':
                        x = x[:-1]

                    step = json.loads(x)
                    if 'stateRoot' in step.keys() and INCLUDE_STATEROOT:
                        steps.append(step)

            except Exception as e:
                logger.info('Exception parsing cpp json:')
                logger.info(e)
                logger.info('problematic line:')
                logger.info(x[:500])

        canon_steps = []

        try:
            for step in steps:
                if 'stateRoot' in step.keys():
                    if len(canon_steps): # dont log state root if no previous EVM steps
                        canon_steps.append(step) # should happen last
                    continue
                if step['op'] in ['INVALID', 'STOP'] :
                    # skip STOPs
                    continue
                if step['op'] not in valid_opcodes:
                    logger.info("got cpp step for an unknown opcode:")
                    logger.info(step)
                    continue

                trace_step = {
                    'pc'  : step['pc'],
                    'gas': '0x{0:01x}'.format(int(step['gas'])) ,
                    'op': opcodes.reverse_opcodes[step['op']],
                    'depth' : step['depth'],
                    'stack' : toHexQuantities(step['stack']),
                }
                canon_steps.append(trace_step)

                # Sometimes, the last one is duplicated. let's just remove that, if so

                if len(canon_steps) > 1:
                    last = canon_steps[-1]
                    slast = canon_steps[-2]
                    if slast['depth'] == last['depth'] and slast['pc'] == last['pc']:
                        canon_steps = canon_steps[:-1]

        except Exception as e:
            logger.info('Exception parsing cpp step:')
            logger.info(e)

        return canon_steps

class PyVM(VM):

    @staticmethod
    def canonicalized(output):
        from . import opcodes

        def formatStackItem(el):
            return '0x{0:01x}'.format(int(el.replace("b", "").replace("'", "")))

        def json_steps():
            for line in output:
                if line.startswith("tx:"):
                    continue
                if line.startswith("tx_decoded:"):
                    continue
                json_index = line.find("{")
                if json_index >= 0:
                    try:
                        yield(json.loads(line[json_index:]))
                    except Exception as e:
                        logger.info("Exception parsing python output:")
                        logger.info(e)
                        logger.info("problematic line:")
                        logger.info(line)
                        yield({})

        canon_steps = []
        for step in json_steps():
            #print (step)
            if 'stateRoot' in step.keys():
                # dont log stateRoot when tx doesnt execute, to match cpp and parity
                if len(canon_steps) and INCLUDE_STATEROOT:
                    canon_steps.append(step)
                continue
            if 'event' not in step.keys():               
                continue
            if step['event'] == 'eth.vm.op.vm':
                if step['op'] not in valid_opcodes:
                    # invalid opcode
                    continue
                if step['op'] == 'STOP':
                    # geth logs code-out-of-range as a STOP, and we 
                    # can't distinguish them from actual STOPs (that pyeth logs)
                    continue

                trace_step = {
                    'opName' : step['op'],
                    'op'     : step['inst'],
                    'depth'  : step['depth'],
                    'pc'     : bstrToInt(step['pc']),
                    'gas'    : bstrToHex(step['gas']),
                }

                trace_step['stack'] = [formatStackItem(el) for el in step['stack']]
                canon_steps.append(trace_step)

        return canon_steps


class GethVM(VM):

    def __init__(self,executable="evmbin", docker = False):
        super().__init__( executable, docker)
        self.genesis_format="geth"

    def makeCommand(self, **kwargs):
        cmd = []

        def get(v, default = None):
            if v in kwargs.keys() and kwargs[v]:
                return kwargs[v]
            return default

        def extend(v, flagname=None, default=None):
            if flagname == None:
                flagname = v

            if get(v, default=default):
                cmd.extend(["--%s" % flagname, str(get(v,default=default))])

        if self.docker: 
            cmd.extend(['docker', 'run', '--rm'])
            # If any files are referenced, they need to be mounted
            if get('genesis') is not None:
                genesis_mount = os.path.dirname(get('genesis'))
                cmd.append('-v')
                if platform.system() == 'Darwin':
                    cmd.append('%s:%s' % (os.path.join('/private', genesis_mount.strip('/')), genesis_mount))
                else:
                    cmd.append('%s:%s' % (genesis_mount, genesis_mount))

        cmd.append( self.executable ) 

        if get('receiver') == "":
            kwargs.pop("receiver", None)
            kwargs['code'] = kwargs.pop('input', "")
            cmd.append("--create")

        extend('code')
        extend('codeFile', 'codefile')
        extend('genesis', 'prestate')
        extend('gas', default=4700000)
        extend('sender')
        extend('receiver')
        extend('input')
        extend('value')

        if not get('memory'):
            cmd.append("--nomemory")
        if get('json'):
            cmd.append("--json")
        if get('statdump'):
            cmd.append("--statdump")
        if get('create'):
            cmd.append("--create")

        cmd.append("run")

        return cmd

    def start(self, **kwargs):
        return self._start(self.makeCommand(**kwargs))

    def execute(self, **kwargs):
        return finishProc(self.start(**kwargs))

    @staticmethod
    def canonicalized(output):
        from . import opcodes
        parsed_steps = []
        for line in output:
            if len(line) > 0 and line[0] == "{":
                try:
                    parsed_steps.append(json.loads(line))
                except Exception as e:
                    logger.warn('Exception [1] parsing geth output:')
                    traceback.print_exc(file=sys.stdout)
                    logger.warn(e)

        canon_steps = []
        if len(parsed_steps) == 0:
            return canon_steps
            
        try:
            if 'output' in parsed_steps[-1]:
                # last one is {"output":"","gasUsed":"0x34a48","time":4787059}
                # Remove    
                parsed_steps = parsed_steps[:-1]

            for step in parsed_steps:
                if 'stateRoot' in step.keys() :
                    # don't log stateRoot when tx doesnt execute, to match cpp and parity
                    # should be last step
                    if len(canon_steps) and INCLUDE_STATEROOT:
                        canon_steps.append(step)
                    
                    continue


                # Ignored for now
                if 'error' in step.keys() and 'output' in step.keys():
                    continue
                if 'time' in step.keys():
                    continue

                if not 'op' in step.keys():
                    logger.warn("Missing 'op': %s" % str(step))
                    continue

                if step['op'] == 0:
                    # skip STOPs
                    continue
                if step['opName'] == "" or step['op'] not in opcodes.opcodes:
                    # invalid opcode
                    continue
                trace_step = {
                    'pc'  : step['pc'],
                    'gas': step['gas'],
                    'op': step['op'],
                    # we want a 0-based depth
                    'depth' : step['depth'] -1,
                    'stack' : step['stack'],
                }
                canon_steps.append(trace_step)
        except Exception as e:
            logger.warn('Exception [2] parsing geth output:')
            traceback.print_exc(file=sys.stdout)
            logger.warn(e)

        return canon_steps



class ParityVM(VM):

    def __init__(self,executable="evmbin", docker = False):
        super().__init__(executable, docker)
    
    def makeCommand(self, **kwargs):
        
        def get(v, default = None):
            if v in kwargs.keys() and kwargs[v]:
                return kwargs[v]
            return default


        code = get('code')
        codeFile = get('codeFile')
        genesis = get('genesis')
        gas = get('gas')
        price = get('price')
        sender = get('sender')
        receiver = get('receiver')
        input = get('input')
        _json = get('json')
        
        if self.docker: 
            cmd = ['docker', 'run','--rm']
            # If any files are referenced, they need to be mounted
            if get('genesis') is not None:
                genesis_mount = os.path.dirname(get('genesis'))
                cmd.append('-v')
                if platform.system() == 'Darwin':
                    cmd.append('%s:%s' % (os.path.join('/private', genesis_mount.strip('/')), genesis_mount))
                else:
                    cmd.append('%s:%s' % (genesis_mount, genesis_mount))

            cmd.append( self.executable )
            cmd.append("/parity-evm")
        else:
            cmd = [self.executable]

        if codeFile is not None :
            with open(codeFile,"r") as f: 
                code = f.read()

        if code is not None:
            cmd.extend(["--code", strip_0x(code)])
        if genesis is not None : 
            cmd.extend(["--chain", genesis])
        if gas is not None: 
            cmd.extend(["--gas","%s" % hex(gas)[2:]])
        if price is not None:
            cmd.extend(["--gas-price","%d" % price] )
        if sender is not None: 
            cmd.extend(["--from", strip_0x(sender)])
        if receiver is not None:
            cmd.extend(["--to",strip_0x(receiver)])
        if input is not None:
            cmd.extend(["--input", input])
        if _json: 
            cmd.append("--json")

        return cmd

    def start(self, **kwargs):
        return self._start(self.makeCommand(**kwargs))


    def execute(self, **kwargs):
        return finishProc(self.start(**kwargs))

    @staticmethod
    def canonicalized(output):
        from . import opcodes
        parsed_steps = []
        for line in output:
            if len(line) > 0 and line[0] == "{":
                try:
                    parsed_steps.append(json.loads(line))
                except Exception as e:
                    logger.warn('Exception [1] parsing parity output:')
                    logger.warn(e)

        canon_steps = []
        try:
            for p_step in parsed_steps:
                if 'test' in p_step.keys():
                    # first step of trace has test name
                    continue

                if 'stateRoot' in p_step.keys():
                    # dont log the stateRoot for basic tx's (that have no EVM steps)
                    # should be last step
                    if len(canon_steps) and INCLUDE_STATEROOT:
                        canon_steps.append(step)
                    continue

                # Ignored for now
                if 'error' in p_step.keys() or 'output' in p_step.keys():
                    continue

                if not 'op' in p_step.keys():
                    logger.warn("Missing 'op': %s" % str(p_step))
                    continue
                    
                if p_step['op'] == 0:
                    # skip STOPs
                    continue
                if p_step['opName'] == "" or p_step['op'] not in opcodes.opcodes:
                    # invalid opcode
                    continue
                trace_step = {
                    'pc'  : p_step['pc'],
                    'gas': p_step['gas'],
                    'op': p_step['op'],
                    # parity depth starts at 1, but we want a 0-based depth
                    'depth' : p_step['depth'] -1,
                    'stack' : p_step['stack'],
                }
                canon_steps.append(trace_step)
        except Exception as e:
            logger.warn('Exception [2] parsing parity output:')
            logger.warn(e)

        return canon_steps






import rlp
from eth_hash.auto import keccak
import binascii

def sha3(seed):
    return keccak(to_string(seed))

def encode_hex(b):
    if isinstance(b, str):
        b = bytes(b, 'utf-8')
    if isinstance(b, (bytes, bytearray)):
        return str(binascii.hexlify(b), 'utf-8')
    raise TypeError('Value must be an instance of str or bytes')

def to_string(value):
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return bytes(value, 'utf-8')
    if isinstance(value, int):
        return bytes(str(value), 'utf-8')

def bytearray_to_bytestr(value):
    return bytes(value)

def decode_hex(s):
    if isinstance(s, str):
        return bytes.fromhex(s)
    if isinstance(s, (bytes, bytearray)):
        return binascii.unhexlify(s)
    raise TypeError('Value must be an instance of str or bytes')

def is_numeric(x): return isinstance(x, int)

def remove_0x_head(s):
    return s[2:] if s[:2] in (b'0x', '0x') else s

def big_endian_to_int(value):
    return int.from_bytes(value, byteorder='big')

def parse_int_or_hex(s):
    if is_numeric(s):
        return s
    elif s[:2] in (b'0x', '0x'):
        s = to_string(s)
        tail = (b'0' if len(s) % 2 else b'') + s[2:]
        return big_endian_to_int(decode_hex(tail))
    else:
        return int(s)

def normalize_address(x, allow_blank=False):
    if is_numeric(x):
        return int_to_addr(x)
    if allow_blank and x in {'', b''}:
        return b''
    if len(x) in (42, 50) and x[:2] in {'0x', b'0x'}:
        x = x[2:]
    if len(x) in (40, 48):
        x = decode_hex(x)
    if len(x) == 24:
        assert len(x) == 24 and sha3(x[:20])[:4] == x[-4:]
        x = x[:20]
    if len(x) != 20:
        raise Exception("Invalid address format: %r" % x)
    return x

def mk_contract_address(sender, nonce):
    return sha3(rlp.encode([normalize_address(sender), nonce]))[12:]



import json

def mktemp(prefix = "", suffix=""):
    import random, string, tempfile
    rand = ''.join([random.choice(string.ascii_letters + string.digits) for n in range(8)])
    temp_path = "%s/%s%s%s" % (tempfile.gettempdir(), prefix, rand, suffix)
    return temp_path

class Genesis(object):
    """ Utility to create genesis files"""

    def __init__(self):
        self.alloc  = {}
        self.coinbase = "0x0000000000000000000000000000000000000000"
        self.timestamp = "0x00"
        self.gasLimit = "0x3D0900"
        self.difficulty = "0x01"
        self.blockNumber = 0
        self.config = {
            "eip150Block": 0, 
            "eip158Block": 0, 
            "eip155Block": 0, 
            "homesteadBlock": 0, 
            "daoForkBlock": 0,
            "byzantiumBlock" : 2000,
        }

    def geth(self):

        g = {
            "nonce":      "0x0000000000000000",
            "difficulty": self.difficulty,
            "mixhash":    "0x0000000000000000000000000000000000000000000000000000000000000000",
            "coinbase": self.coinbase,
            "timestamp": self.timestamp,
            "number":    "0x{:02x}".format(self.blockNumber),
            "parentHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
            "extraData":  "0x0000000000000000000000000000000000000000000000000000000000000000",
            "gasLimit": self.gasLimit,
            "alloc": self.alloc,
            "config": self.config,

        }
        return g

    def parity(self):

        builtins = {
                "0000000000000000000000000000000000000001": { "builtin": 
                    { "name": "ecrecover", "pricing": { "linear": { "base": 3000, "word": 0 } } } },
                "0000000000000000000000000000000000000002": { "builtin": 
                    { "name": "sha256", "pricing": { "linear": { "base": 60, "word": 12 } } } },
                "0000000000000000000000000000000000000003": { "builtin": 
                    { "name": "ripemd160", "pricing": { "linear": { "base": 600, "word": 120 } } } },
                "0000000000000000000000000000000000000004": { "builtin": 
                    { "name": "identity", "pricing": { "linear": { "base": 15, "word": 3 } } } },
                "0000000000000000000000000000000000000005": { "builtin": {"activate_at": self.config['byzantiumBlock'], "name": "modexp", "pricing": { "modexp": { "divisor": 20 }}}},
                "0000000000000000000000000000000000000006": { "builtin": { "activate_at": self.config['byzantiumBlock'], "name": "alt_bn128_add",  "pricing": { "linear": { "base": 500, "word": 0 }}}},
                "0000000000000000000000000000000000000007": { "builtin": { "activate_at": self.config['byzantiumBlock'], "name": "alt_bn128_mul",  "pricing": { "linear": { "base": 40000, "word": 0 }}}},
                "0000000000000000000000000000000000000008": { "builtin": { "activate_at": self.config['byzantiumBlock'], "name": "alt_bn128_pairing", "pricing": { "alt_bn128_pairing": { "base": 100000, "pair": 80000 }}}},
            }
        builtins.update(self.alloc)
        g = {
            "name": "lab",
            "engine": {
               "Ethash": {
                  "params": {
                    "minimumDifficulty": "0x020000",
                    "difficultyBoundDivisor": "0x0800",
                    "durationLimit": "0x0d",
                    "blockReward": "0x4563918244F40000",
                    "registrar": "",
                    "frontierCompatibilityModeLimit": "0x0",
                    "eip150Transition"   : 0,
                    "eip155Transition"   : 0,
                    "eip160Transition"   : 0,
                    "eip161abcTransition": 0,
                    "eip161dTransition"  : 0,
                    "eip649Reward":"0x29A2241AF62C0000",
                    "eip649Transition" : self.config['byzantiumBlock'],
                    "eip100bTransition": self.config['byzantiumBlock'],

                    }
                }
            },
            "genesis": {
                "seal": {
                  "ethereum": {
                    "nonce": "0x0000000000000042",
                    "mixHash": "0x0000000000000000000000000000000000000000000000000000000000000000"
                  }
                },
                "difficulty": "0x400",
                "author": "0x3333333333333333333333333333333333333333",
                "timestamp": "0x0",
                "parentHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
                "extraData": "0x0",
                "gasLimit": "0x8000000",
            },
            "params": {
                "maxCodeSize": 24576,
                "gasLimitBoundDivisor": "0x0400",
                "accountStartNonce": "0x0",
                "maximumExtraDataSize": "0x20",
                "minGasLimit": "0x1388",
                "networkID" : "0x0",
                "eip98Transition": "0x0",
                "eip86Transition": "0x0",
                "eip140Transition" : self.config['byzantiumBlock'],
                "eip211Transition" : self.config['byzantiumBlock'],
                "eip214Transition" : self.config['byzantiumBlock'],
                "eip658Transition" : self.config['byzantiumBlock'],


    # Also new pre
            },
            "accounts": builtins,
        }
        return g

    def has(self, account):
        return account.lower() in self.alloc.keys()

    def setCoinbase(self, coinbase):
        self.coinbase = coinbase

    def setGasLimit(self, gasLimit):
        self.gasLimit = gasLimit

    def setTimestamp(self, timestamp):
        self.timestamp = timestamp

    def setDifficulty(self, difficulty):
        self.difficulty = difficulty

    def setBlockNumber(self, blockNumber):
        self.blockNumber = int(blockNumber, 16)

    def setConfigHomestead(self):
        self.config['byzantiumBlock'] = 2000
        self.config['eip158Block'] = 2000
        self.config['eip155Block'] = 2000
        self.config['eip150Block'] = 2000
        self.config['homesteadBlock'] = 0

    def setConfigTangerineWhistle(self):
        self.config['byzantiumBlock'] = 2000
        self.config['eip158Block'] = 2000
        self.config['eip155Block'] = 2000
        self.config['eip150Block'] = 0
        self.config['homesteadBlock'] = 0

    def setConfigSpuriousDragon(self):
        self.config['byzantiumBlock'] = 2000
        self.config['eip158Block'] = 0
        self.config['eip155Block'] = 0
        self.config['eip150Block'] = 0
        self.config['homesteadBlock'] = 0

    def setConfigMetropolis(self):
        self.config['byzantiumBlock'] = 0
        self.config['eip158Block'] = 0
        self.config['eip155Block'] = 0
        self.config['eip150Block'] = 0
        self.config['homesteadBlock'] = 0

    def addPrestateAccount(self, account):
        self.alloc[account['address'].lower()] = {
            "balance" : account['balance'],
            "code" : account['code'],
            "nonce" : account['nonce'],
        }
        if 'storage' in account:
            self.alloc[account['address'].lower()]['storage'] = {}
            for key in account['storage']:
                self.alloc[account['address'].lower()]['storage'][key] = account['storage'][key]

    def add(self, account): 
            
        """
        Data format from EtherChain:
                {
                "address": "0x6090a6e47849629b7245dfa1ca21d94cd15878ef",
                "balance": 0,
                "nonce": null,
                "code": "0x60...",
                "name": null,
                "storage": null,
                "firstSeen": "2017-04-26T19:12:56.000Z"
            }
            """
        n = account['nonce'] 
        if n is None:
            n = 0
        b ="0x%x" % (account['balance'])
        code = account['code'].hex() # code is a HexBytes object
        self.alloc[account['address'].lower()] = {
            "balance" : b, 
            "code" : code, 
            "nonce" : hex(n), 
        }

    def codeAt(self,addr):
        addr = addr.lower()
        if addr in self.alloc.keys():
            acc = self.alloc[addr]
            if 'code' in acc:
                return acc['code']
        return ""

    def addStorage(self, account, key, value):
        ac = self.alloc[account.lower()]
        key = "0x{:064x}".format(int(key,16))

        if 'storage' not in ac.keys():
            ac['storage'] = {}

        if 'hex' in dir(value):
            value = value.hex()
        else:
            value = "0x{:064x}".format(int(value,16)) 

        ac['storage'][key]=value


    def export(self,prefix="genesis"):
        geth_genesis = self.export_geth(prefix="%s-genesis-geth_" % prefix)
        parity_genesis = self.export_parity(prefix="%s-genesis-parity_" % prefix)
        
        return (geth_genesis, parity_genesis)

    def export_geth(self, prefix = None):
        temp_path = mktemp(prefix = prefix, suffix=".json")
        with open(temp_path, 'w') as f :
            json.dump(self.geth(),f)
        return temp_path

    def export_parity(self, prefix = None):
        temp_path = mktemp(prefix = prefix, suffix=".json")
        with open(temp_path, 'w') as f :
            json.dump(self.parity(),f)
        return temp_path

    def prettyprint(self):
        import pprint
        pp = pprint.PrettyPrinter(indent=2)
        pp.pprint(self.geth())



#!/usr/bin/env python3
import re

from evmlab.opcodes import parseCode

"""
Solidity source code mappings, as 
documented [here](http://solidity.readthedocs.io/en/develop/miscellaneous.html#source-mappings)

"""


def update(original, changes):
    retval = []
    for i in range(0, len(original)):
        val = original[i]
        if i < len(changes) and len(changes[i]) > 0:
            val = changes[i]
        retval.append(val)
    return retval


def parseSourceMap(maptext):
    mapping = []

    if maptext is None:
        return mapping

    entries = maptext.split(";")
    m = ["", "", "", ""]
    for e in entries:
        vals = e.split(":")
        m = update(m, vals)
        mapping.append(m)
    return mapping


class Contract():
    _create = False

    bin = None
    ins = None
    binRuntime = None
    insRuntime = None
    lastSource = None
    name = ""

    def __init__(self, sources, contract=None, name=""):
        self.sources = sources or []
        self._contractTexts = {}
        self._sourceCache = {}
        self.name = name 

        self._loadContract(contract)

    @property
    def create(self):
        return self._create

    @create.setter
    def create(self, val):
        self._create = val
        self._loadContractTexts()

    @property
    def contractTexts(self):
        if len(self._contractTexts.keys()) == 0:
            self._loadContractTexts()

        return self._contractTexts

    def isInitialized(self):
        return self.bin is not None or self.binRuntime is not None

    def getSourceCode(self, pc):
        try:
            [s, l, f, j] = self._getInstructionMapping(pc)
            f = int(f)
            c = self.contractTexts[f]
        except KeyError:
            if self.lastSource:
                return self._sourceCache[self.lastSource][0], self._sourceCache[self.lastSource][1]
            
            return "Missing code", (0, 0)
        s = int(s)
        l = int(l)

        h = hash((s, l, f, j))
        if h in self._sourceCache:
            self.lastSource = h
            return self._sourceCache[h][0], self._sourceCache[h][1]

        # contract is missing, return the last valid ins mapping
        if f < 0:
            while True:
                pc -= 1
                try:
                    [s, l, f, j] = self._getInstructionMapping(pc)
                except KeyError:
                    f = -1
                f = int(f)
                if f > 0:
                    c = self.contractTexts[f]
                    s = int(s)
                    l = int(l)
                    break

        # see if text contains multiple contracts
        contract_start_indices = [m.start(0)
                                  for m in re.finditer('^ *contract ', c)]

        # for multi contract files, get the start of the contract for the current instruction
        if len(contract_start_indices) > 1:
            contract_start = 0
            contract_end = -1

            for i in contract_start_indices:
                if i == s:
                    contract_start = s
                    break
                elif i > s:
                    # get the previous index
                    ci = contract_start_indices.index(i) - 1
                    if ci >= 0:
                        contract_start = contract_start_indices[ci]
                    break
                elif s > i and i == contract_start_indices[-1]:
                    contract_start = contract_start_indices[-1]

            pos = contract_start + c[contract_start:].find('{')
            openBr = 0
            while pos < len(c):
                if c[pos] == '{':
                    openBr += 1
                elif c[pos] == '}':
                    openBr -= 1

                if openBr == 0:
                    contract_end = pos + 1
                    break

                pos += 1

            # return only the contract we're interested in
            # we need to update the bytes start & end pos to reflect the truncated text we are returning
            res = (c[contract_start:contract_end], [s - contract_start, l])
            self._sourceCache[h] = res

            self.lastSource = h
            return res[0], res[1]
        else:
            self._sourceCache[h] = (c, [s, l])
            self.lastSource = h
            return c, [s, l]

    def _getInstructionMapping(self, pc):
        """
        :param pc: programCounter to fetch mapping for
        :return: [s, l, f, j] where:
             s is the byte-offset to the start of the range in the source file,
             l is the length of the source range in bytes and
             f is the source index mentioned above.
             j can be either i, o or -
               i   signifying whether a jump instruction goes into a function,
               o   returns from a function or
               -   is a regular jump as part of e.g. a loop.
        """
        i = self._getMappingIndex(pc)
        mapping = self.mapping if self.create else self.mappingRuntime

        return mapping[i]

    def _getMappingIndex(self, pc):
        ins = self.ins if self.create else self.insRuntime
        pcs = list(ins.keys())

        if pc in pcs:
            return pcs.index(pc)

        raise KeyError

    def _loadContract(self, contract):
        if not contract:
            return

        def load(key):
            try:
                return contract[key]
            except KeyError:
                print("Contract JSON missing key: %s" % key)
                return None

        bytecode = load('bin-runtime')
        if bytecode:
            self.binRuntime = bytecode
            self.insRuntime = parseCode(bytecode)

        bytecode = load('bin')
        if bytecode:
            self.bin = bytecode
            self.ins = parseCode(bytecode)

        self.mappingRuntime = parseSourceMap(load('srcmap-runtime'))
        self.mapping = parseSourceMap(load('srcmap'))

    def _loadContractTexts(self):
        self._sourceCache = {}
        mapping = self.mapping if self.create else self.mappingRuntime

        contract_indexes = set()
        for [s, l, f, j] in mapping:
            f = int(f)
            if (f > 0):
                contract_indexes.add(f)

        for i in contract_indexes:
            self._contractTexts[i] = self.sources[i]


#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
Opviewer - Ethereum Transaction Debugger
"""
import os
import json
import argparse, math
import logging

from evmlab.context import buildContexts
from evmlab.contract import Contract
from evmlab import reproduce, utils
from evmlab import vm as VMUtils
from evmlab.opcodes import reverse_opcodes

logger = logging.getLogger(__name__)

try:
    import ethereum_input_decoder
except ImportError:
    ethereum_input_decoder = None
    logger.warning(
        "Transaction input decoder package not installed. run `#> pip install evmlab[abidecoder]` to install.")
try:
    import urwid
except ImportError:
    logger.fatal("Console UI (urwid) package not installed. run `#> pip install evmlab[consolegui]` to install.")

# Python3 support
# Todo: six instead of custom checks? do we support py2?
try:
    xrange
except NameError:
    xrange = range


class SourceException(Exception):
    # TODO: create exception hierarchy in evmlab package
    pass


class Console:

    @staticmethod
    def bold(text):
        # Todo: not used? get rid of this or use it
        return '\033[1m{}\033[0m'.format(text)


# Todo: get rid of this
# OUTPUT_DIR = "%s/output/" % os.path.dirname(os.path.realpath(__file__))  # should not use realpath of file as we are inside the package.


# NEWLINE_PATTERN = re.compile('\n')


class DebugViewer(object):
    # todo: review staticmethods and check if they really belong to debugviewer (visualization vs. generic helper)

    def __init__(self):

        self.memptr = 0
        self.stackptr = 0
        self.opptr = 0
        self.srcptr = 0
        self.srctrack = True
        self.snapn = 0

        self.ops_view = None
        self.mem_view = None
        self.memref_view = None
        self.stack_view = None
        self.trace_view = None
        self.source_view = None
        self.help_view = None

    @staticmethod
    def wrap(x, y):
        return urwid.LineBox(x, y)

    @staticmethod
    def getMemoryReference(opcode):

        # opcodes with memory references:
        # CALLDATACOPY, CODECOPY, EXTCODECOPY, RETURNDATACOPY, MLOAD, MSTORE*
        # LOG*, CREATE, CALL, CALLCODE, RETURN, DELEGATECALL, STATICCALL, REVERT
        # format of mem_opcodes dict is [[mem_offset, data_size], default]

        # TODO: use mnemonics instead of opcodes
        # TODO: move to class
        mem_opcodes = {
            0x37: [0, 2],
            0x39: [0, 2],
            0x3c: [1, 3],
            0x3e: [0, 2],
            0x51: [0, -1, 32],
            0x52: [0, -1, 32],
            0x53: [0, -1, 1],
            0xa0: [0, 1],
            0xa1: [0, 1],
            0xa2: [0, 1],
            0xa3: [0, 1],
            0xa4: [0, 1],
            0xf0: [1, 2],
            0xf1: [3, 4],
            0xf2: [3, 4],
            0xf3: [0, 1],
            0xf4: [2, 3],
            0xfa: [2, 3],
            0xfd: [0, 1]
        }

        if opcode in mem_opcodes.keys():
            e = mem_opcodes[opcode]
            return e
        else:
            return -1

    @staticmethod
    def memRefResolve(mem, refs, st, msg, opname, oplimit):
        append = ""
        st = st[::-1]
        mem = mem[2:]
        mrefs = [0, 0]
        mrefs[0] = int(st[refs[0]], 16)
        if (refs[1] == -1):
            mrefs[1] = refs[2]
        else:
            mrefs[1] = int(st[refs[1]], 16)
        if oplimit > 0 and mrefs[1] > oplimit:
            mrefs[1] = oplimit
            append = "..."
        if mrefs[1] < 0 or mrefs[0] < 0 or mrefs[0] > len(mem) / 2:
            mrefs = [0, 0]
            append = " - memory access beyond expansion"
        if (mrefs[0] + mrefs[1]) > len(mem) / 2:
            mrefs[1] = len(mem / 2) - mrefs[0]
            append = " - attempted read beyond memory bound of %d bytes" % (mrefs[0] + mrefs[1] - len(mem / 2))
        return msg + " " + opname + " memory ref:\n" + "0x" + "".join(
            mem[(mrefs[0] * 2):(mrefs[0] + mrefs[1]) * 2]) + append + "\n"

    @staticmethod
    def dumpArea(f, name, content):
        # TODO: get rid of this?
        f.write(name + "\n\n")
        f.write(content + "\n\n")

    @staticmethod
    def opSource(c, opcode, srcptr, track=True, length=12):
        """
        @brief formats an execution step to the corresponding source code
        """

        if 'pc' in opcode.keys():
            contract, code_pos = c.getSourceCode(opcode['pc'])

            if code_pos == [-1, -1]:
                raise SourceException()

            start = code_pos[0]
            end = start + code_pos[1]
            code = contract[start:end]

            # determine which line the code is on
            start_code_line = contract[:start].count('\n')
            # start_code_line = len(re.findall(NEWLINE_PATTERN, contract[:start]))
            # num of lines of code
            code_lines = code.count('\n') + 1
            # code_lines = len(re.findall(NEWLINE_PATTERN, code)) + 1

            split = contract.splitlines()

            if track:
                if start_code_line == 0 or code_lines >= length:
                    srcptr = start_code_line
                    view = split[start_code_line: start_code_line + length]
                else:
                    offset = math.floor((length - code_lines) / 2)
                    line_end = min(start_code_line + code_lines + offset, len(split))
                    line_start = max(line_end - length, 0)

                    srcptr = line_start
                    view = split[line_start: line_end]
            else:
                if srcptr + length > len(split):
                    srcptr = max(len(split) - length, 0)
                view = split[srcptr: srcptr + length]

            viewtxt = "\n".join(view)

            # highlight the current code in the view

            # determine the code start offset for the view
            code_start_offset = start - len("\n".join(split[:srcptr])) - 1
            code_end_offset = code_start_offset + code_pos[1]
            if (code_start_offset <= 0):
                # current code starts before the view
                if (code_end_offset < 0):
                    # current code is before the view
                    pass
                elif (code_end_offset > len(viewtxt)):
                    # entire view is current code
                    viewtxt = [('source', viewtxt)]
                else:
                    # first part of view is current code
                    viewtxt = [('source', viewtxt[:code_end_offset]), viewtxt[code_end_offset:]]
            else:
                # current code starts within or after the view
                if (code_start_offset > len(viewtxt)):
                    # current code is after the view
                    pass
                elif (code_end_offset > len(viewtxt)):
                    # second part of view is current code
                    viewtxt = [viewtxt[:code_start_offset], ('source', viewtxt[code_start_offset:])]
                else:
                    # current code is within the current view
                    viewtxt = [viewtxt[:code_start_offset], ('source', viewtxt[code_start_offset:code_end_offset]),
                               viewtxt[code_end_offset:]]

            return srcptr, viewtxt
        else:
            return srcptr, ""

    @staticmethod
    def getStackAnnotations(opcode):
        """
        Returns a list of annotations for stack elements, given the opcode 'op'.
        Example, ifcode = CALL:
        ['gas', 'address','value', 'instart', 'insize', 'outstart', 'outsize']
        """
        # TODO: impmort opcodes and reuse them
        opcodes = {
            0x00: ['STOP', 0, 0, 0, []],
            0x01: ['ADD', 2, 1, 3, ['operand', 'operand']],
            0x02: ['MUL', 2, 1, 5, ['operand', 'operand']],
            0x03: ['SUB', 2, 1, 3, ['operand', 'operand']],
            0x04: ['DIV', 2, 1, 5, ['numerator', 'denominator']],
            0x05: ['SDIV', 2, 1, 5, ['numerator', 'denominator']],
            0x06: ['MOD', 2, 1, 5, ['x', 'modulator']],
            0x07: ['SMOD', 2, 1, 5, ['x', 'modulator']],
            0x08: ['ADDMOD', 3, 1, 8, ['operand', 'operand', 'modulator']],
            0x09: ['MULMOD', 3, 1, 8, ['operand', 'operand', 'modulator']],
            0x0a: ['EXP', 2, 1, 10, ['base', 'exponent']],
            0x0b: ['SIGNEXTEND', 2, 1, 5, ['byte', 'bit']],
            0x10: ['LT', 2, 1, 3, ['x', 'y']],
            0x11: ['GT', 2, 1, 3, ['x', 'y']],
            0x12: ['SLT', 2, 1, 3, ['x', 'y']],
            0x13: ['SGT', 2, 1, 3, ['x', 'y']],
            0x14: ['EQ', 2, 1, 3, ['x', 'y']],
            0x15: ['ISZERO', 1, 1, 3, ['x']],
            0x16: ['AND', 2, 1, 3, ['x', 'y']],
            0x17: ['OR', 2, 1, 3, ['x', 'y']],
            0x18: ['XOR', 2, 1, 3, ['x', 'y']],
            0x19: ['NOT', 1, 1, 3, ['x']],
            0x1a: ['BYTE', 2, 1, 3, ['index', 'word']],
            0x20: ['SHA3', 2, 1, 30, ['offset', 'size']],
            0x30: ['ADDRESS', 0, 1, 2, []],
            0x31: ['BALANCE', 1, 1, 20, ['address']],
            0x32: ['ORIGIN', 0, 1, 2, []],
            0x33: ['CALLER', 0, 1, 2, []],
            0x34: ['CALLVALUE', 0, 1, 2, []],
            0x35: ['CALLDATALOAD', 1, 1, 3, ['position']],
            0x36: ['CALLDATASIZE', 0, 1, 2, []],
            0x37: ['CALLDATACOPY', 3, 0, 3, ['memOffset', 'dataOffset', 'length']],
            0x38: ['CODESIZE', 0, 1, 2, []],
            0x39: ['CODECOPY', 3, 0, 3, ['memOffset', 'codeOffset', 'length']],
            0x3a: ['GASPRICE', 0, 1, 2, []],
            0x3b: ['EXTCODESIZE', 1, 1, 20, ['address']],
            0x3c: ['EXTCODECOPY', 4, 0, 20, ['address', 'memOffset', 'codeOffset', 'length']],
            0x3d: ['RETURNDATASIZE', 0, 1, 2, []],
            0x3e: ['RETURNDATACOPY', 3, 0, 3, ['memOffset', 'dataOffset', 'length']],
            0x40: ['BLOCKHASH', 1, 1, 20, ['blocknum']],
            0x41: ['COINBASE', 0, 1, 2, []],
            0x42: ['TIMESTAMP', 0, 1, 2, []],
            0x43: ['NUMBER', 0, 1, 2, []],
            0x44: ['DIFFICULTY', 0, 1, 2, []],
            0x45: ['GASLIMIT', 0, 1, 2, []],
            0x50: ['POP', 1, 0, 2, ['popvalue']],
            0x51: ['MLOAD', 1, 1, 3, ['offset']],
            0x52: ['MSTORE', 2, 0, 3, ['mempos', 'data']],
            0x53: ['MSTORE8', 2, 0, 3, ['mempos', 'data']],
            0x54: ['SLOAD', 1, 1, 50, ['location']],
            0x55: ['SSTORE', 2, 0, 0, ['location', 'data']],
            0x56: ['JUMP', 1, 0, 8, ['destination']],
            0x57: ['JUMPI', 2, 0, 10, ['destination', 'cond']],
            0x58: ['PC', 0, 1, 2, []],
            0x59: ['MSIZE', 0, 1, 2, []],
            0x5a: ['GAS', 0, 1, 2, []],
            0x5b: ['JUMPDEST', 0, 0, 1, []],
            0xa0: ['LOG0', 2, 0, 375, ['memstart', 'memsize']],
            0xa1: ['LOG1', 3, 0, 750, ['memstart', 'memsize', 'topic1']],
            0xa2: ['LOG2', 4, 0, 1125, ['memstart', 'memsize', 'topic1', 'topic2']],
            0xa3: ['LOG3', 5, 0, 1500, ['memstart', 'memsize', 'topic1', 'topic2', 'topic3']],
            0xa4: ['LOG4', 6, 0, 1875, ['memstart', 'memsize', 'topic1', 'topic2', 'topic3', 'topic4']],
            0xf0: ['CREATE', 3, 1, 32000, ['value', 'mstart', 'msize']],
            0xf1: ['CALL', 7, 1, 40, ['gas', 'address', 'value', 'instart', 'insize', 'outstart', 'outsize']],
            0xf2: ['CALLCODE', 7, 1, 40, ['gas', 'address', 'value', 'instart', 'insize', 'outstart', 'outsize']],
            0xf3: ['RETURN', 2, 0, 0, ['memstart', 'length']],
            0xf4: ['DELEGATECALL', 6, 0, 40, ['gas', 'address', 'instart', 'insize', 'outstart', 'outsize']],
            0xfa: ['STATICCALL', 6, 1, 40, ['gas', 'address', 'instart', 'insize', 'outstart', 'outsize']],
            0xfd: ['REVERT', 2, 0, 0, ['memstart', 'length']],
            0xff: ['SUICIDE', 1, 0, 0, ['beneficiary']],
        }

        dup_annotation = ['duptarget']
        swap_annotation = ['operand', 'operand']
        for i in range(1, 17):
            opcodes[0x7f + i] = ['DUP' + str(i), i, i + 1, 3, dup_annotation]
            opcodes[0x8f + i] = ['SWAP' + str(i), i + 1, i + 1, 3, swap_annotation]
            dup_annotation = [''] + dup_annotation
            swap_annotation = ['swaptarget'] + [''] * i + ['swaptarget']

        if opcode in opcodes.keys():
            return opcodes[opcode][4]

        return []

    @staticmethod
    def hexdump(src, length=16, sep='.', minrows=8, start=0, prevsrc=""):
        """
        @brief Return {src} in hex dump.
        """
        txt = lambda c: chr(c) if 0x20 <= c < 0x7F else "."

        result = []
        result.append('           00 01 02 03 04 05 06 07  08 09 0A 0B 0C 0D 0E 0F  |  [ --ascii--]')
        result.append('')
        rows = []

        for i in xrange(0, 16):
            subSrc = src[2 * (start * 16 + i * 16):2 * (start * 16 + i * 16) + length * 2]
            hexa = ''
            text = ''
            if len(subSrc) > 0:
                for h in xrange(0, len(subSrc), 2):
                    if h == length:
                        hexa += ' '
                    byte = int(subSrc[h:h + 2], 16)

                    # Check if it changed from op before
                    changed = False
                    if prevsrc is not None:
                        index = 2 * (start + i) * 16 + h
                        if index + 2 > len(prevsrc):
                            changed = True
                        elif int(prevsrc[index:index + 2], 16) != byte:
                            changed = True

                    if changed:
                        hexa += "{:02x}•".format(byte)
                    else:
                        hexa += "{:02x} ".format(byte)
                    text += txt(byte)

            rows.append('{:08x}:  {:<49} | {:<16} '.format(16 * (start + i), hexa, text))
            if len(rows) == minrows:
                break
        result.extend(rows)
        return '\n'.join(result)

    @staticmethod
    def stackdump(st, length=16, minrows=8, start=0, opcode=None):
        """
        @brief Return the 'st' stackdump nicely formatted
        Also adds stack annotations if 'opcode' is provided
        """

        result = []
        annotations = []
        if opcode is not None:
            annotations = DebugViewer.getStackAnnotations(opcode)

        top_item = start + min(minrows, len(st) - 1)  # 0
        bottom_item = start  # 0

        # Note:
        # Stack 'st' is [bottom , ... , top]
        # Annotation 'annotations' is [top , .. , bottom]

        for i in xrange(start, min(minrows, len(st))):
            stackvalue = ""
            annotation = ""

            if i < len(st):
                val = int(st[-1 - i], 16)
                stackvalue = "  {:02} : {:.>64x}".format(i, val)
            else:
                stackvalue = "  {:02} : {:64}".format(i, "")

            if len(annotations) > i:
                annotation = annotations[i]

            result.append("{} {}".format(stackvalue, annotation))

        return '\n'.join(result)

    @staticmethod
    def opDump(obj, addr):
        """
        @brief returns information about current exeution step
        """

        def attr(x, tryCastToInt=True):
            prelim = ('bold', "{:10}: ".format(x))
            if x in obj.keys():
                v = obj[x]
                if type(v) != type(0) and tryCastToInt:
                    try:
                        v = int(v, 16)
                    except Exception as e:
                        # It's text
                        return [prelim, "{:6}\n".format(v.strip())]
                # An int
                if type(v) == type(0):
                    return [prelim, "{:6d} (0x{:02x})\n".format(v, v)]
                return [prelim, "{:6} \n".format(v)]
            else:
                # Not available
                return [prelim, "<N/A>\n"]

        result = []

        result.append(attr('pc'))
        result.append(attr('op'))
        result.append(attr('opName', False))
        result.append(attr('gasCost'))
        result.append(attr('gas'))
        result.append(attr('memSize'))
        result.append(attr('depth'))
        if addr:
            result.append([('bold', "{:10}: ".format("addr")), addr])

        return result

    @staticmethod
    def toText(op):
        """
        @brief Formats an execution-step to one line of text
        """

        def attr(x, fmt, tryCastToInt=True):
            if x in op.keys():
                v = op[x]
                if type(v) != type(0) and tryCastToInt:
                    try:
                        v = int(v, 16)
                    except Exception as e:
                        pass
                return fmt.format(v)

            return "N/A"

        result = []

        def precheck(op):
            if not op.keys():
                return "END"
            if 'stateRoot' in op.keys():
                return "stateRoot {}".format(op['stateRoot'])
            if 'error' in op.keys() and op['error'] not in [None, '']:
                if str(op['error']).find('gas') > -1:
                    return "OOG"
                return "Err: {}".format(op['error'])
            return None

        abort = precheck(op)

        if abort:
            result.append(abort)
            return " ".join(result)

        result.append(attr('pc', '{:>5d}'))

        result.append(attr('opName', '{:>12}', False))
        result.append(attr('op', '{:>5x}'))
        result.append(attr('gas', '{:>8x}'))
        result.append(attr('gasCost', '{:>8d}'))
        result.append(attr('depth', '{:>2d}'))
        return " ".join(result)

    @staticmethod
    def opTrace(ops=[], sel=0, offset=0):
        """
        @brief formats a list of instructions to a table
        """
        header = "|".join(["step    ",
                           " pc   ",
                           "  opname  ",
                           " opcode ",
                           " gas  ",
                           "gascost",
                           "depth"
                           ])
        result = [header, ""]

        for i, op in enumerate(ops):
            if i + offset == sel:
                result.append("{:<4} >> {}".format(i + offset, DebugViewer.toText(op)))
            else:
                result.append("{:<4}    {}".format(i + offset, DebugViewer.toText(op)))

        return "\n".join(result)

    def setTrace(self, trace, op_contracts=[], txhash=None, txinput=None):
        self.operations = trace
        self.op_contracts = op_contracts

        ops_view = urwid.Text(self.getOp())
        mem_view = urwid.Text(self.getMem())
        memref_view = urwid.Text(self.getMemref())
        stack_view = urwid.Text(self.getStack())
        trace_view = urwid.Text(self.getTrace())
        source_view = urwid.Text(self.getSource())
        help_view = urwid.Text(self.getHelp())
        # palettes are currently not used
        palette = [
            ('banner', 'black', 'light gray'),
            ('streak', 'black', 'dark red'),
            ('bg', 'black', 'dark blue'), ]
        palette = [
            ('headings', 'white,underline', 'black', 'bold,underline'),  # bold text in monochrome mode
            ('body_text', 'dark cyan', 'light gray'),
            ('buttons', 'yellow', 'dark green', 'standout'),
            ('section_text', 'body_text'),  # alias to body_text
            ('source', 'white', 'black', 'bold'),  # bold text in monochrome mode
        ]

        self.ops_view = ops_view
        self.mem_view = mem_view
        self.memref_view = memref_view
        self.stack_view = stack_view
        self.trace_view = trace_view
        self.source_view = source_view
        self.help_view = help_view

        # indicate the online lookup with a * at the end of decoded
        try:
            txinput_decoded = ethereum_input_decoder.AbiMethod.from_input_lookup(ethereum_input_decoder.Utils.str_to_bytes(txinput)) if ethereum_input_decoder else "<input decoder not installed>"
        except Exception as e:  # not going to import from eth_abi to not make the code depending on it
            txinput_decoded = "!! DecodingError: %s" %e


        inp_view = urwid.Text("""
  > tx:       %s
  > input:    %s
  > decoded*: %s""" % (txhash,
                      txinput,
                      txinput_decoded))

        top = DebugViewer.wrap(urwid.Pile([
            DebugViewer.wrap(inp_view, ""),
            urwid.Columns([
                urwid.Pile([
                    DebugViewer.wrap(ops_view, "Op"),
                    DebugViewer.wrap(trace_view, "Trace")]
                ),
                urwid.Pile([
                    DebugViewer.wrap(mem_view, "Memory"),
                    DebugViewer.wrap(memref_view, "Memory Reference by Opcode"),
                    DebugViewer.wrap(stack_view, "Stack"),
                    DebugViewer.wrap(source_view, "Source"),
                ])
            ])
        ]), "Retromix")

        horiz = urwid.Pile([top, DebugViewer.wrap(help_view, "Help")])
        fill = urwid.Filler(horiz, 'top')

        # self.dbg("Loaded %d operations" % len(self.operations) )

        loop = urwid.MainLoop(fill, palette, unhandled_input=lambda input: self.show_or_exit(input))
        loop.run()

    def _op(self, key=None, default=None):

        if self.opptr > len(self.operations) - 1:
            return default

        op = self.operations[self.opptr]
        if key == None:
            return op
        if key not in op.keys():
            return default
        if op[key]:
            return op[key]
        return default

    def _prevop(self, key=None, default=None):

        if self.opptr > len(self.operations) - 2:
            return default

        op = self.operations[self.opptr - 1]
        if key == None:
            return op
        if key not in op.keys():
            return default
        if op[key] is not None:
            return op[key]
        return default

    def getOp(self):
        addr = None
        if (len(self.op_contracts) > 0):
            c = self.op_contracts[self.opptr]
            addr = "{} ({})".format(c.address, c.name.split(':')[-1])

        return DebugViewer.opDump(self._op(default={'pc': 1}), addr)

    def getMem(self):
        m = self._op('memory', "0x")
        if type(m) is list:
            m = "0x%s" % "".join(m)
        prev_m = self._prevop('memory', "0x")
        if type(prev_m) is list:
            prev_m = "0x%s" % "".join(prev_m)
        return DebugViewer.hexdump(m[2:], start=self.memptr, prevsrc=prev_m[2:])

    def _getMemref(self, bound):
        m = self._op('memory', [])
        mc = ""
        mc_prev = ""
        ms = DebugViewer.getMemoryReference(self._op('op', "0"))
        ms_prev = DebugViewer.getMemoryReference(self._prevop('op', "0"))
        if type(ms) is list:
            mc = DebugViewer.memRefResolve(m, ms, self._op('stack', []), "Pre-exec", self._op('opName', 'op'), bound)
        if type(ms_prev) is list:
            mc_prev = DebugViewer.memRefResolve(m, ms_prev, self._prevop('stack', []), "Post-exec",
                                                self._prevop('opName', 'op'), bound)
        return mc + mc_prev

    def getMemref(self):
        return self._getMemref(256)

    def getStack(self):
        st = self._op('stack', [])
        opcode = self._op('op', None)
        return DebugViewer.stackdump(st, start=self.stackptr, opcode=opcode)

    def getTrace(self):
        # Trace 2 * pad + 1 lines

        pad = 12
        sel = self.opptr
        start = max(sel - pad, 0)

        end = min(start + 2 * pad + 1, len(self.operations) - 1)

        ops = self.operations[start: end]
        return DebugViewer.opTrace(ops=ops, sel=sel, offset=start)

    def getSource(self, track=None):

        if len(self.op_contracts) == 0:
            return "No Source Provided"

        op = self.operations[self.opptr]

        if 'depth' not in op:
            return ""

        if not self.op_contracts[self.opptr]:
            return "Missing context"

        if track is None:
            track = self.srctrack

        try:
            self.srcptr, txt = DebugViewer.opSource(self.op_contracts[self.opptr], op, self.srcptr, track=track)
            return txt
        except SourceException:
            return self.source_view.text

    def getHelp(self):
        return """Key navigation
        a: Trace up        s: Mem up     d: Stack up    f: Source up    t: track source on/off    m: write data to snapshot file
        z: Trace down      x: Mem down   c: Stack down  v: Source down  Use uppercase for large steps
    press `q` to quit
        """

    def _refresh(self):
        self.source_view.set_text(self.getSource())  # needs to occur before getOp to print correct addr
        self.ops_view.set_text(self.getOp())
        self.trace_view.set_text(self.getTrace())
        self.mem_view.set_text(self.getMem())
        self.memref_view.set_text(self.getMemref())
        self.stack_view.set_text(self.getStack())
        self.help_view.set_text(self.getHelp())

    def dbg(self, text):
        if self.help_view is not None:
            self.help_view.set_text(text)

    def show_or_exit(self, key):
        """
        @brief handles key-events
        """
        if key in ('q', 'Q'):
            raise urwid.ExitMainLoop()

        step = 1
        if key in ('A', 'Z'):
            step = 10 if len(self.operations) <= 1000 else 100
        elif key in ('S', 'X', 'D', 'C', 'F', 'V'):
            step = 10

        # UP trace
        if key in ('a', 'A'):
            self.opptr = max(0, self.opptr - step)
            self._refresh()

        # DOWN trace
        if key in ('z', 'Z'):
            self.opptr = min(self.opptr + step, len(self.operations) - 1)
            self._refresh()

        # UP mem
        if key in ('s', 'S'):
            self.memptr = max(0, self.memptr - step)
            self.mem_view.set_text(self.getMem())

        # DOWN mem
        if key in ('x', 'X'):
            self.memptr = self.memptr + step
            self.mem_view.set_text(self.getMem())

        # UP stack
        if key in ('d', 'D'):
            self.stackptr = max(0, self.stackptr - step)
            self.stack_view.set_text(self.getStack())

        # DOWN stack
        if key in ('c', 'C'):
            self.stackptr = self.stackptr + step
            self.stack_view.set_text(self.getStack())

        # UP source
        if key in ('f', 'F'):
            self.srcptr = max(0, self.srcptr - step)
            self.source_view.set_text(self.getSource(track=False))

        # DOWN source
        if key in ('v', 'V'):
            self.srcptr = self.srcptr + step
            self.source_view.set_text(self.getSource(track=False))

        if key in ('t', 'T'):
            self.srctrack = not self.srctrack
            if self.srctrack:
                self.srcptr = 0
            self.source_view.set_text(self.getSource())

        if key in ('g', 'G'):
            self.dbg("TODO: Implement GOTO")

        if key in ('m', 'M'):
            snap_name = os.path.join(os.getcwd(), "evmlab.state.snapshot%s.txt" % self.snapn)
            ops = "".join(str(e[0][1] + e[1]) for e in self.getOp())
            try:
                sf = open(snap_name, "w")
                DebugViewer.dumpArea(sf, "OP STATE", str(ops))
                DebugViewer.dumpArea(sf, "TRACE", self.getTrace())
                DebugViewer.dumpArea(sf, "MEMORY REFERENCE", self._getMemref(0))
                DebugViewer.dumpArea(sf, "MEMORY DUMP", self.getMem())
                DebugViewer.dumpArea(sf, "STACK", self.getStack())
                sf.close()
                self.dbg("Saved snapshot to %s" % snap_name)
            except Exception as e:
                self.dbg(str(e))
            self.snapn += 1


class EvmTrace(object):
    """
    Main Wrapper class to handle Evm Traces
    """

    def __init__(self, api="https://mainnet.infura.io/remix"):

        self.api = utils.getApi(api)
        self.source_path = None

        self.txhash = None
        self.txinput = None

        # internal state
        self.ops = []
        self.contracts = []
        self.op_contracts = []

    @staticmethod
    def get_evm_handler(vmtype, path, docker=True):
        """
        utility function to get Geth/Parity VMObjects

        :param vmtype: string type of evm geth|parity
        :param path: path to evm binary
        :param docker: is this a docker container
        :return: either GethVM or ParityVM object
        """
        if vmtype == "geth":
            vmclass = VMUtils.GethVM
        elif vmtype == "parity":
            vmclass = VMUtils.ParityVM
        else:
            raise Exception("vmtype not supported.")

        return vmclass(executable=path, docker=docker)

    def show(self):
        """
        show() urwid ui

        @info: ideally this is not part of this class. we should probably move this to another class e.g. UrwidUI(EvmTrace)
        :return:
        """
        if not self.ops:
            raise Exception("need to reproduce/load trace first")

        DebugViewer().setTrace(self.ops, self.op_contracts, self.txhash, self.txinput)

    def reproduce(self, tx, vm):
        """

        :param tx:  # tx hash
        :param vm:  # see OpViewer.get_evm_handler(...)
        :return: self
        """
        artefacts, vm_args = reproduce.reproduceTx(tx, vm, self.api)
        logger.debug("done reproducing transaction trace...")
        ## TODO: fix artefacts
        ## TODO: get rid of saveFiles
        ## TODO: generally get rid of tempfile logic.
        # saved_files = utils.saveFiles(OUTPUT_DIR, artefacts)
        # fname = os.path.join(saved_files['json-trace']['path'], saved_files['json-trace']['name'])

        return self.load_trace(path=artefacts['json-trace'])

    def load_trace(self, tx=None, _json=None, path=None):
        if _json:
            return self.load_trace_json(data=_json)
        elif tx:
            self.txhash = tx
            self.txinput = self.api.getTransaction(tx)["input"]
            return self.load_trace(_json=self.api.traceTransaction(tx=tx))
        elif path:
            return self.load_trace_file(path=path)
        raise Exception("either tx, _json debugtrace or path to a json_debugtrace required")

    def load_trace_json(self, data):
        """Parse the output from debug_traceTransaction"""
        ops = []

        if 'jsonrpc' in data:
            # get rid of jsonrpc envelope if it is available
            data = data['result']
        xops = data['structLogs']
        for op in xops:
            newOp = dict(op)  # get rid of attributeDict if data comes from api
            newOp['opName'] = op['op']
            newOp['op'] = reverse_opcodes[op['op']]

            if 'memory' in op.keys():
                if op['memory'] is None:
                    newOp['memory'] = "0x"
                else:
                    newOp['memory'] = "0x" + "".join(op['memory'])
                    logger.debug("Memory set to %s" % newOp['memory'])
            ops.append(newOp)
        logger.debug("Loaded %d items from structlogs" % len(ops))

        self.ops = ops
        logger.debug("trace loaded (structLogs).")
        return self

    def load_trace_file(self, path):
        if not os.path.isfile(path):
            raise Exception("%s - is not a file" % path)

        logger.debug("loading trace file: %s" % path)
        with open(path) as f:
            #
            # 1) try json tracefile
            #
            try:
                return self.load_trace_json(json.load(f))
            except json.decoder.JSONDecodeError:
                logger.debug('Failed to parse file in debug_traceTransaction format')
            #
            # 2) try line-by-line json
            #
            logger.debug("Trying to load %s in geth format")
            f.seek(0)  # rewind
            data = f.read()
            try:
                return self._loadJsonObjects(data)
            except Exception as e:
                logger.debug(e)
            #
            # 3) try load weird json format
            #
            try:
                return self._loadWeirdJson(data)
            except Exception as e:
                logger.debug(e)

        raise Exception("could not load json trace file")

    def _loadJsonObjects(self, data):
        """Load the json from geth `evm`"""

        ops = []
        for l in data.split("\n"):
            l = l.strip()
            if l.startswith("#"):
                continue
            if len(l) == 0:
                continue
            op = json.loads(l)
            if 'action' in op.keys():
                continue
            ops.append(op)

        if not ops:
            raise Exception("could not load json line-by-line code")

        self.ops = ops
        logger.debug("trace loaded (json objects)")
        return self

    def _loadWeirdJson(self, data):
        ops = []

        text = ""
        for l in data.split("\n"):
            # However, sometimes they're spread out
            if l.strip() == '{':
                text = ""
            text = text + l.strip()
            if l.strip() == '}':
                jsondata = json.loads(text)
                ops.append(jsondata)

        if not ops:
            raise Exception("could not load weird json")

        self.ops = ops
        logger.debug("trace loaded (weird json).")
        return self

    def load_source(self, text, bytecode):
        # TODO: load single contract source
        pass

    def load_contract_sources_from_combined_json(self, tx, combined_json, source_prefix=''):
        # TODO: untested, need reference file
        contracts = []
        sources = []

        # get sources (text) for all contracts
        for file in [os.path.join(source_prefix, s) for s in combined_json['sourceList']]:
            with open(file) as s:
                logger.debug(s.name)
                sources.append(s.read())

        # get contract
        for contract, val in combined_json['contracts'].iteritems():
            contracts.append(Contract(sources, val, contract))

        self.contracts = contracts
        self.op_contracts = buildContexts(self.ops, self.api, contracts, tx)
        return self

    def save(self, path):
        # TODO: save artefacts, trace files, ...
        pass


def main():
    """
    ##python3 opviewer.py -f test.json

    :return:
    """

    description = """
Tool to explore a json-trace in a debug-like interface
"""
    examples = """
# Generate a trace and analyze with sources

python3 opviewer.py -s /path/to/contracts -j /path/to/combined.json --hash txHash

# Analyse a trace

python3 opviewer.py -f example.json

# Analyse a trace with sources

python3 opviewer.py -f example.json -s /path/to/contracts -j /path/to/combined.json --hash txHash
"""

    parser = argparse.ArgumentParser(description=description, epilog=examples,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-v', "--verbose", action="store_true", default=False, help="Set loglevel to DEBUG")

    evmchoice = parser.add_mutually_exclusive_group()
    evmchoice.add_argument("-f", "--file", type=str, help="Trace file to load")
    evmchoice.add_argument('-g', '--geth-evm', type=str,
                           help="Geth EVM binary or docker image name", default=None)
    evmchoice.add_argument('-p', '--parity-evm', type=str, default=None,
                           help="Parity EVM binary or docker image name")

    parser.add_argument("--no-docker", action="store_true",
                        help="Set to true if using a local binary instead of a docker image")

    sourcesettings = parser.add_argument_group('Sources', 'Settings used to display contract sources')
    sourcesettings.add_argument("-s", "--source", type=str, default="./", help="Contract source code directory")
    sourcesettings.add_argument("-j", "--json", type=str, help="Compiler combined-json output")
    sourcesettings.add_argument('--hash', type=str, help="hash of the tx to view")

    web3settings = parser.add_argument_group('Web3',
                                             'Settings about where to fetch information from when displaying contract sources (default infura)')
    web3settings.add_argument("--web3", type=str, default="https://mainnet.infura.io/remix",
                              help="Web3 API url to fetch info from (default 'https://mainnet.infura.io/remix'")

    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    if args.hash and not args.hash.startswith("0x"):
        args.hash = "0x%s" % args.hash

    if args.parity_evm and args.geth_evm:
        parser.error("please either specify --geth-evm or --parity-evm and not both.")

    # -- End of arg normalization

    logger.debug("--start--")
    # TODO: talk to infura to get a debug_traceTransaction enabled endpoint (be fair, not reuse remix ;))
    trace = EvmTrace(api=args.web3)

    logger.debug("init done.")

    if not args.file:
        # reproduce or get trace from remote api
        if not args.hash:
            parser.error('hash is required ')

        # reproduce
        if args.parity_evm or args.geth_evm:
            logger.debug("reproducing trace ...")
            vmtype, vmpath = ("parity", args.parity_evm) if args.parity_evm else ("geth", args.geth_evm)
            logger.debug("using VM: %s @ %s (docker=%s)" % (vmtype, vmpath, not args.no_docker))
            vm = EvmTrace.get_evm_handler(vmtype=vmtype, path=vmpath, docker=not args.no_docker)
            trace.reproduce(tx=args.hash, vm=vm)
        else:
            # load from api
            logger.debug("fetching traces from remote api ...")
            trace.load_trace(tx=args.hash)
    else:
        # load from file
        trace.load_trace(path=args.file)

        # eventually load combined.json
        if args.json:
            if not args.hash:
                parser.error('hash is required ')

            with open(args.json, 'r') as f:
                combined = json.load(f)
            trace.load_contract_sources_from_combined_json(tx=args.hash,
                                                           combined_json=combined, source_prefix=args.source)
    trace.show()
    logger.debug("--end--")


if __name__ == '__main__':
    logging.basicConfig(format='[%(filename)s - %(funcName)20s() ][%(levelname)8s] %(message)s',
                        level=logging.INFO)
    main()




import os
import requests
import json

here = os.path.dirname(os.path.abspath(__file__))

def log(x):
    print("[+] %s" % x)


def savejson(fname, obj):
    fqfn = '%s/datastore/%s' % (here, fname)
    with open(fqfn, 'w+') as outfile:
        outfile.write(json.dumps(obj))

    log("Wrote file %s" % fqfn)

def loadjson(fname):
#   return None
    try:
        with open('%s/datastore/%s' % (here, fname), 'r') as infile:
            return json.loads(infile.read())
    except Exception as e:
        print(e)
        return None


class EtherChainAPI():

    def __init__(self, offline = False):
        self.offline = offline
        self.cachedBlocks = None
        self.flushcounter = 0

    def getBlockInfo(self, blockNumberOrHash):
        """
        Fields :
            {"number":101010
            ,"hash":"0x3e5c2f13bc01a0a32c7eda3560ec3ede19c9d2f6c0fda8436999eee2914e862d"
            ,"parentHash":"0x8290679004240a002dcf274bf56310c2654b5137c643e5c20ff647ebda61033b"
            ,"uncleHash":"0x1dcc4de8dec75d7aab85b567b6ccd41ad312451b948a7413f0a142fd40d49347"
            ,"coinbase":"0xc603ab5ef41fc61aa8b76beddb09e3cfcbb174bd"
            ,"root":"0x3df3a270c940336f7585d9b62a37fa04b2e2794f3267915f65b202e37a93c4e5"
            ,"txHash":"0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421"
            ,"difficulty":3871451743585
            ,"gasLimit":3141592
            ,"gasUsed":0
            ,"time":"2015-08-17T12:53:09.000Z"
            ,"extra":"0x476574682f76312e302e312f77696e646f77732f676f312e342e32"
            ,"mixDigest":null
            ,"nonce":"0xeaedaa0e2baa6334"
            ,"tx_count":0
            ,"uncle_count":0
            ,"size":544
            ,"blockTime":1
            ,"reward":5000000000000000000}      
        """
        # Load cached data
        if self.cachedBlocks == None:
            self.cachedBlocks = loadjson(".blocks.json") or {}

        if(self.cachedBlocks.has_key(blockNumberOrHash)):
            return self.cachedBlocks[blockNumberOrHash]


        url = "https://etherchain.org/api/block/%s"  % str(blockNumberOrHash)
        log(url)
        data = requests.get(url).json()['data'][0]

        #save
        self.cachedBlocks[data['hash']] = data
        self.cachedBlocks[data['number']] = data

        self.flushcounter = self.flushcounter+1
        if self.flushcounter > 50:#Always on now
            savejson('.blocks.json', self.cachedBlocks)
            self.flushcounter =0 

        return data
    def getBlockTime(self, blockNumberOrHash):
        return self.getBlockInfo(blockNumberOrHash)['time']


    def getTransaction(self, txhash):
        return requests.get("https://etherchain.org/api/tx/%s" % txhash).json()['data'][0]

    def getAllTransactions(self, address):
        """ Returns an iterator of all transactions to/from the specified address
            Internally, this shelves transactions to disk, and only queries for new 
            transactions
        """
        # Load cached data
        cachedTransactions = loadjson(".%s-transactions.json" % address) or {'byhash':{}}

        tx_byhash = cachedTransactions['byhash']

        stats = {"added" : 0}
        
        def insert(tx):
            if not tx_byhash.has_key(tx['hash']):
                tx_byhash[tx['hash']] = tx
                stats['added'] = stats['added']+1
                return True
            return False

        #Fetch new data
        abort = self.offline
        while not abort:
            offset = len(tx_byhash)

            url = "https://etherchain.org/api/account/%s/tx_asc/%d" % (address, offset)
            log(url)
            resp = requests.get(url)
            if resp.status_code == 200 :
                jsondata = resp.json()['data']
                for tx in jsondata:
                    if not insert(tx): 
                        log("Error, we got a tx we already have. Probably missing earlier transactions for %s" % address)
                        abort = True

                if len(jsondata) < 50:
                    abort = True

            else:
                raise Exception("Failed to get %s history" % instr)


        #Save cache
        if stats['added'] > 0:
            savejson(".%s-transactions.json" % address, cachedTransactions)
            print("Added %d transactions for %s" % (stats['added'], address))

        # Return txs starting at lowest nonce
        for h,tx in tx_byhash.items():
            yield tx


    def outgoingTransactions(self, address):
        for tx in self.getAllTransactions(address):
            if tx['sender'].lower() == address.lower():
                yield tx

    def contractCreateTransactions(self, address):
        for tx in self.outgoingTransactions(address):
            if tx['newContract'] != 0:
                yield tx

    def contractCreationTransaction(self, address):
        for tx in self.incomingTransactions(address):
            if tx['newContract'] != 0:
                yield tx
        return


    def incomingTransactions(self, address):
        for tx in self.getAllTransactions(address):
            if tx['recipient'].lower() == address.lower():
                yield tx

    def sendersTo(self, address):
        address = address.lower()
        allTxs = []
        totalCost =0
        senders = set()
        
        log("Loading senders to %s" % address)
        
        for tx in self.incomingTransactions(address):
            gasUsed = float(tx['gasUsed'])
            price = float(tx['price'])
            cost = (gasUsed * price)
            senders.add(tx['sender'])
            allTxs.append(tx['time'])
            #print(tx)
            totalCost = totalCost + cost

        allTxs.sort()
        return (senders, totalCost, allTxs)

    def getAccount(self, address):
        return requests.get("https://etherchain.org/api/account/%s" % address).json()['data'][0]


    def getBalance(self, address):
        return requests.get("https://etherchain.org/api/account/%s" % address).json()['data'][0]['balance']

    def getCode(self, address):
        return requests.get("https://etherchain.org/api/account/%s" % address).json()['data'][0]['code']


    def getBalances(self,addresses):
        return [(address,self.getBalance(address)) for address in addresses]


    def findNewContracts(self, address):
        """ Finds all contracts created by a given address
        Returns all previously unseen contract"""

        foundContracts = {}

        # Find them all
        for tx in self.contractCreateTransactions(address):
            newC = tx['recipient'].lower()
            foundContracts[newC] = tx

        #Check which we already know about
        knownContracts = loadjson(".%s-contracts.json" % address) or {}

        diff = []
        for k,v in foundContracts.iteritems():
            if not k in knownContracts.keys():
                diff.append(v)
                knownContracts[k] = v
        # Save
        if len(diff) > 0:
            savejson(".%s-contracts.json" % address, knownContracts)
        return diff

    def getAllContracts(self, address):
        """ Finds all contracts created by a given address"""
        self.findNewContracts(address)
        knownContracts = loadjson(".%s-contracts.json" % address) or {}
        return knownContracts.values()


if __name__ == '__main__':
    api = EtherChainAPI(offline=True)
    



import shelve, traceback
from . import utils

class MultiApi(object):

    """ Helper class for using several API:s. 
    Currently supports web3 and etherchain

    web3 is a bit better, since it allows for querying about balance
    and things at specified block height.
    """

    def __init__(self, web3 = None, etherchain = None):
        self.web3 = web3
        self.etherchain = etherchain

    def _getCached(self,key):
        db = shelve.open(".api_cache")
        obj = None
        if key in db:
            obj = db[key]
        db.close()
        return obj

    def _putCached(self,key, obj):
        db = shelve.open(".api_cache")
        db[key] = obj
        db.close()

    def getAccountInfo(self, address, blnum = None):
        acc = {}

        
        print("GetAccountInfo(%s, %s)"% (address, str(blnum)))

        if blnum is not None: 
            cachekey = "%s-%d" % (address, blnum)
            cached = self._getCached(cachekey)
            if cached is not None:
                return cached

        if self.web3 is not None:
            # web3 only accepts checksummed addresses
            chk_address = utils.checksumAddress(address) 
            acc['balance'] = self.web3.eth.getBalance(chk_address, blnum)
            acc['code']    = self.web3.eth.getCode(chk_address, blnum)
            acc['nonce']   = self.web3.eth.getTransactionCount(chk_address, blnum)
            acc['address'] = address
    
            # testrpc will return 0x0 if no code, geth expects 0x
            if acc['code'] == '0x0':
                acc['code'] = '0x'

            # cache it, but only if it's at a specific block number
            if blnum is not None:
                self._putCached(cachekey, acc)

        elif self.etherchain is not None: 
            acc = self.etherchain.getAccount(address)

        return acc

    def getTransaction(self,h):

        cachekey = "tx-%s" % h

        o = self._getCached(cachekey)
        if o is not None:
            return o

        translations = [("sender", "from"),
                        ("recipient", "to"),
                        ("block_id", "blockNumber" )]

        if self.web3 : 
            obj = self.web3.eth.getTransaction(h)
            obj_dict = {}
            for a in obj:
              obj_dict[a] = obj[a]
            for (a,b) in translations:
                obj_dict[a] = obj_dict[b]

        else:
            obj = self.etherchain.getTransaction(h)
            obj_dict = {key: value for (key, value) in obj}
            for (a,b) in translations:
                obj_dict[b] = obj_dict[a]

        self._putCached( cachekey, obj_dict)
        return obj_dict

    def getStorageSlot(self, addr, key, blnum = None):

        print("GetStorageSlot(%s, %s, %s)"% (addr, key, str(blnum)))

        if blnum is not None: 
            cachekey = "%s-%d-%s" % (addr,blnum,key)
            cached = self._getCached(cachekey)
            if cached is not None:
                return cached

        if self.web3:
            try:
                value = self.web3.eth.getStorageAt(addr, key, blnum)
                self._putCached(cachekey, value)
                return value
            except Exception as e:
                print("ERROR OCCURRED: trace may not be correct")    
                traceback.print_exc()
                return ""
            
        else:
            print("getStorageSlot not implemented for etherchain api")
            return ""

    def traceTransaction(self, tx, disableStorage=False, disableMemory=False, disableStack=False, tracer=None,
                               timeout=None):
        if self.web3 is None:
            raise Exception("debug_traceTransaction requires web3 to be configured")

        # TODO: caching
        return self.web3.manager.request_blocking("debug_traceTransaction",
                                                  [tx, {"disableStorage": disableStorage,
                                                        "disableMemory": disableMemory,
                                                        "disableStack": disableStack,
                                                        "tracer": tracer,
                                                        "timeout": timeout}])



from copy import copy
import collections
import binascii

from . import parse_int_or_hex,decode_hex,remove_0x_head,bytearray_to_bytestr,encode_hex

# Taken from https://github.com/ethereum/pyethereum/blob/develop/ethereum/opcodes.py
# Done this way to reduce dependencies a bit
# schema: [opcode, ins, outs, gas]

opcodes = {
    0x00: ['STOP', 0, 0, 0],
    0x01: ['ADD', 2, 1, 3],
    0x02: ['MUL', 2, 1, 5],
    0x03: ['SUB', 2, 1, 3],
    0x04: ['DIV', 2, 1, 5],
    0x05: ['SDIV', 2, 1, 5],
    0x06: ['MOD', 2, 1, 5],
    0x07: ['SMOD', 2, 1, 5],
    0x08: ['ADDMOD', 3, 1, 8],
    0x09: ['MULMOD', 3, 1, 8],
    0x0a: ['EXP', 2, 1, 10],
    0x0b: ['SIGNEXTEND', 2, 1, 5],
    0x10: ['LT', 2, 1, 3],
    0x11: ['GT', 2, 1, 3],
    0x12: ['SLT', 2, 1, 3],
    0x13: ['SGT', 2, 1, 3],
    0x14: ['EQ', 2, 1, 3],
    0x15: ['ISZERO', 1, 1, 3],
    0x16: ['AND', 2, 1, 3],
    0x17: ['OR', 2, 1, 3],
    0x18: ['XOR', 2, 1, 3],
    0x19: ['NOT', 1, 1, 3],
    0x1a: ['BYTE', 2, 1, 3],
    0x20: ['SHA3', 2, 1, 30],
    0x30: ['ADDRESS', 0, 1, 2],
    0x31: ['BALANCE', 1, 1, 20],
    0x32: ['ORIGIN', 0, 1, 2],
    0x33: ['CALLER', 0, 1, 2],
    0x34: ['CALLVALUE', 0, 1, 2],
    0x35: ['CALLDATALOAD', 1, 1, 3],
    0x36: ['CALLDATASIZE', 0, 1, 2],
    0x37: ['CALLDATACOPY', 3, 0, 3],
    0x38: ['CODESIZE', 0, 1, 2],
    0x39: ['CODECOPY', 3, 0, 3],
    0x3a: ['GASPRICE', 0, 1, 2],
    0x3b: ['EXTCODESIZE', 1, 1, 20],
    0x3c: ['EXTCODECOPY', 4, 0, 20],
    0x3d: ['RETURNDATASIZE', 0, 1, 2],
    0x3e: ['RETURNDATACOPY', 3, 0, 3],
    0x40: ['BLOCKHASH', 1, 1, 20],
    0x41: ['COINBASE', 0, 1, 2],
    0x42: ['TIMESTAMP', 0, 1, 2],
    0x43: ['NUMBER', 0, 1, 2],
    0x44: ['DIFFICULTY', 0, 1, 2],
    0x45: ['GASLIMIT', 0, 1, 2],
    0x50: ['POP', 1, 0, 2],
    0x51: ['MLOAD', 1, 1, 3],
    0x52: ['MSTORE', 2, 0, 3],
    0x53: ['MSTORE8', 2, 0, 3],
    0x54: ['SLOAD', 1, 1, 50],
    0x55: ['SSTORE', 2, 0, 0],
    0x56: ['JUMP', 1, 0, 8],
    0x57: ['JUMPI', 2, 0, 10],
    0x58: ['PC', 0, 1, 2],
    0x59: ['MSIZE', 0, 1, 2],
    0x5a: ['GAS', 0, 1, 2],
    0x5b: ['JUMPDEST', 0, 0, 1],
    0xa0: ['LOG0', 2, 0, 375],
    0xa1: ['LOG1', 3, 0, 750],
    0xa2: ['LOG2', 4, 0, 1125],
    0xa3: ['LOG3', 5, 0, 1500],
    0xa4: ['LOG4', 6, 0, 1875],
    0xf0: ['CREATE', 3, 1, 32000],
    0xf1: ['CALL', 7, 1, 40],
    0xf2: ['CALLCODE', 7, 1, 40],
    0xf3: ['RETURN', 2, 0, 0],
    0xf4: ['DELEGATECALL', 6, 0, 40],
    0xfa: ['STATICCALL', 6, 1, 40],
    0xfd: ['REVERT', 2, 0, 0],
    0xff: ['SUICIDE', 1, 0, 0],
}

opcodesMetropolis = { 0x3d, 0x3e, 0xfa, 0xfd }

for i in range(1, 33):
    opcodes[0x5f + i] = ['PUSH' + str(i), 0, 1, 3]

for i in range(1, 17):
    opcodes[0x7f + i] = ['DUP' + str(i), i, i + 1, 3]
    opcodes[0x8f + i] = ['SWAP' + str(i), i + 1, i + 1, 3]

reverse_opcodes = {}
for o in opcodes:
    vars()[opcodes[o][0]] = opcodes[o]
    reverse_opcodes[opcodes[o][0]] = o

# Non-opcode gas prices
GDEFAULT = 1
GMEMORY = 3
GQUADRATICMEMDENOM = 512  # 1 gas per 512 quadwords
GSTORAGEREFUND = 15000
GSTORAGEKILL = 5000
GSTORAGEMOD = 5000
GSTORAGEADD = 20000
GEXPONENTBYTE = 10    # cost of EXP exponent per byte
GCOPY = 3             # cost to copy one 32 byte word
GCONTRACTBYTE = 200   # one byte of code in contract creation
GCALLVALUETRANSFER = 9000   # non-zero-valued call
GLOGBYTE = 8          # cost of a byte of logdata

GTXCOST = 21000       # TX BASE GAS COST
GTXDATAZERO = 4       # TX DATA ZERO BYTE GAS COST
GTXDATANONZERO = 68   # TX DATA NON ZERO BYTE GAS COST
GSHA3WORD = 6         # Cost of SHA3 per word
GSHA256BASE = 60      # Base c of SHA256
GSHA256WORD = 12      # Cost of SHA256 per word
GRIPEMD160BASE = 600  # Base cost of RIPEMD160
GRIPEMD160WORD = 120  # Cost of RIPEMD160 per word
GIDENTITYBASE = 15    # Base cost of indentity
GIDENTITYWORD = 3     # Cost of identity per word
GECRECOVER = 3000     # Cost of ecrecover op

GSTIPEND = 2300

GCALLNEWACCOUNT = 25000
GSUICIDEREFUND = 24000

# Anti-DoS HF changes
SLOAD_SUPPLEMENTAL_GAS = 150
CALL_SUPPLEMENTAL_GAS = 660
EXTCODELOAD_SUPPLEMENTAL_GAS = 680
BALANCE_SUPPLEMENTAL_GAS = 380
CALL_CHILD_LIMIT_NUM = 63
CALL_CHILD_LIMIT_DENOM = 64
SUICIDE_SUPPLEMENTAL_GAS = 5000


def parseCode(code):
    code = code[2:] if code[:2] == '0x' else code

    try:
        codes = [c for c in decode_hex(code)]
    except ValueError as e:
        print(code)
        raise Exception("Did you forget to link any libraries?") from e

    instructions = collections.OrderedDict()
    pc = 0
    length = None
    while pc < len(codes):
        try:
            opcode = opcodes[codes[pc]]
        except KeyError:
            opcode = ['INVALID', 0, 0, 0]
        if opcode[0][:4] == 'PUSH':
            opcode = copy(opcode)
            length = codes[pc] - 0x5f
            pushData = codes[pc + 1 : pc + length + 1]
            pushData = "0x" + encode_hex(bytearray_to_bytestr(pushData))
            if type(pushData) is not str:
                pushData = pushData.decode()
            opcode.append(pushData)

        instructions[pc] = opcode

        if length is not None:
            pc += length
            length = None
        pc += 1

    return instructions


#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
EVMLab command line utility: #> python3 -m evmlab <subcommand> <args>
"""
import sys
import logging

from .tools import opviewer
from .tools import reproducer


def usage(msg=""):
    description = """
EVMLab command line utility

Usage:
    #> python3 -m evmlab <subcommand> <args>

    available subommands
    * opviewer      ...     reproduce and debug transactions
    * reproducer    ...     reproduce transactions


    %s
""" % msg
    print(description)
    return 1  # exitcode / hint error condition


# configure available subcommands here
SUBCOMMAND = {'opviewer': lambda: opviewer.main(),
              'reproducer': lambda: reproducer.main()}


def main():
    if not len(sys.argv)>1:
        sys.exit(usage("- subcommand is mandatory"))

    cmd_util = sys.argv.pop(1)  # pop <util> from cmdline args

    ret = SUBCOMMAND.get(cmd_util.strip().lower(), usage)()
    sys.exit(ret)


if __name__ == "__main__":
    logging.basicConfig(format='[%(filename)s - %(funcName)20s() ][%(levelname)8s] %(message)s',
                        level=logging.INFO)
    main()



import os, shutil
from web3 import Web3
from . import etherchain
from . import multiapi

def getApi(url):
    web3 = Web3(Web3.HTTPProvider(url, request_kwargs={'timeout': 60}))
    chain = etherchain.EtherChainAPI()
    return multiapi.MultiApi(web3 = web3, etherchain = chain)

def checksumAddress(lcAddress):
    return Web3.toChecksumAddress(lcAddress)


def saveFiles(destination, artefacts):
    """
    Copies the supplied artefacts to the right folder

    TODO: Add option to save files into a zip file for download instead
    """

    print("Saving files")
    print("")
    saved = {}

    for desc, path in artefacts.items():
        if os.path.isfile(path):
            fname = os.path.basename(path)
            shutil.copy(path, destination)
            saved[desc] = {'path': destination,'name': fname}
            print("* %s -> %s%s" % (desc, destination, fname) )
        else:
            print("Failed to save %s - not a file" % path)

    return saved



"""
Move reusable utility code that can be made available via cmdline at python -m evmlab <tool> as a module to this package

example:
    see opviewer
"""
from . import opviewer
from .reproducer import reproducer

__ALL__ = ['opviewer', 'reproducer']


#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
import sys, os
import argparse
import re
import zipfile
import tempfile
import logging

from evmlab import reproduce, utils
from evmlab import vm as VMUtils

logger = logging.getLogger(__name__)

try:
    import flask
    app = flask.Flask(__name__, template_folder=os.path.join(os.path.dirname(__file__), 'templates'))
    logger.info("Flask init: template_folder: %s" % os.path.join(os.path.dirname(__file__), 'templates'))
except ImportError:
    logger.warning("Flask not installed, disabling web mode")
    app = None

OUTPUT_DIR = tempfile.mkdtemp(prefix="evmlab")


def create_zip_archive(input_files, output_archive):
    """
    Bundles artefacts into a zip-file

    @param input_artefacts - map of files to save
    @param output_archive prefix to zip-file name
    """
    logger.debug("creating zip archive %s for input artefacts:")
    zipf = zipfile.ZipFile(output_archive, 'w', zipfile.ZIP_DEFLATED)
    for location, name in input_files:
        logger.debug("adding %s as %s to archive..." % (location, name))
        zipf.write(location, name)
    zipf.close()


if app:
    hash_regexp = re.compile("0x[0-9a-f]{64}")


    @app.route("/")
    def test():
        logger.debug("rendering index...")
        return flask.render_template("index.html")


    @app.route('/reproduce/<txhash>')
    def reproduce_tx(txhash):
        logger.debug("reproducing transaction %s" % txhash)

        # Verify input
        if not hash_regexp.match(txhash):
            logger.debug("rendering index(invalid tx hash)...")
            return flask.render_template("index.html", message="Invalid tx hash")

        try:
            artefacts, vm_args = reproduce.reproduceTx(txhash, app.vm, app.api)
            logger.debug("done reproducing transaction trace...")
        except Exception as e:
            logger.exception("exception thrown while reproducing transaction...")
            return flask.render_template("index.html", message=str(e))

        logger.debug("saving artefacts to %s" % OUTPUT_DIR)
        saved_files = utils.saveFiles(OUTPUT_DIR, artefacts)

        # Some tricks to get the right command for local replay
        p_gen = saved_files['parity genesis']['name']
        g_gen = saved_files['geth genesis']['name']
        vm_args['genesis'] = g_gen
        command = app.vm.makeCommand(**vm_args)
        logger.debug("vm command: %s" % command)

        logger.debug("creating zip archive for artefacts")
        prefix = txhash[:8]
        output_archive = os.path.join(OUTPUT_DIR, "%s.zip" % prefix)

        # create a list of files to pack with zipFiles
        input_files = [(os.path.join(v['path'], v['name']), v['name']) for v in saved_files]

        create_zip_archive(input_files=input_files, output_archive=output_archive)

        logger.debug("rendering reproduce_tx...")
        return flask.render_template("index.html",
                                     files=saved_files, zipfile="%s.zip" % prefix,
                                     message="Transaction tracing seems to have been successfull. Use the following command to execute locally",
                                     code=" \\\n\t".join(command))


    @app.route('/download/<path:filename>')
    def download_file(filename):
        logger.debug("rendering download_file...")
        return flask.send_from_directory(OUTPUT_DIR, filename, as_attachment=True)


def test(vm, api):
    print("Doing tests")
    # Jumpdest-analysis attack
    tx = ""
    # Parity-wallet attack
    tx = "0x9dbf0326a03a2a3719c27be4fa69aacc9857fd231a8d9dcaede4bb083def75ec"
    # tenx token transfer (should include SLOADS)
    tx = "0xd6d519043d40691a36c9e718e47110309590e6f47084ac0ec00b53718e449fd3"

    return reproduce.reproduceTx(tx, vm, api)


def main():
    description = """
Tool to reproduce on-chain events locally. 
This can run either as a command-line tool, or as a webapp using a built-in flask interface.
    """
    examples = """
Examples

# Reproduce a tx with a local evm binary
python3 reproducer.py --no-docker -g ~/go/src/github.com/ethereum/go-ethereum/build/bin/evm --hash 0xd6d519043d40691a36c9e718e47110309590e6f47084ac0ec00b53718e449fd3 

# Reproduce a tx with a docker evm
python3 reproducer.py -g holiman/gethvm --hash 0xd6d519043d40691a36c9e718e47110309590e6f47084ac0ec00b53718e449fd3

# Start the reproducer webapp using the default geth docker image: 
python3 reproducer.py -w localhost

Unfinished: 

* This does not _quite_ work with parity, yet, because parity does not load the code in genesis for the 'to'
  -account, it expects the code to be given as an argument. 

    """
    parser = argparse.ArgumentParser(description=description, epilog=examples,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    evmchoice = parser.add_mutually_exclusive_group()
    evmchoice.add_argument('-g', '--geth-evm', type=str,
                           help="Geth EVM binary or docker image name", default="holiman/gethvm")
    evmchoice.add_argument('-p', '--parity-evm', type=str, default=None,
                           help="Parity EVM binary or docker image name")

    parser.add_argument("--no-docker", action="store_true",
                        help="Set to true if using a local binary instead of a docker image")

    web_or_direct = parser.add_mutually_exclusive_group()
    web_or_direct.add_argument('-x', '--hash', type=str,
                               help="Don't run webapp, just lookup hash")
    if app:
        web_or_direct.add_argument('-w', '--www', type=str, help="Run webapp on given interface (interface:port)")
        parser.add_argument('-d', '--debug', action="store_true", default=False,
                            help="Run flask in debug mode (WARNING: debug on in production is insecure)")

    parser.add_argument('-t', '--test', action="store_true", default=False,
                        help="Dont run webapp, only local tests")

    web3settings = parser.add_argument_group('Web3', 'Settings about where to fetch information from (default infura)')
    web3settings.add_argument("--web3", type=str, default="https://mainnet.infura.io/",
                              help="Web3 API url to fetch info from (default 'https://mainnet.infura.io/'")

    args = parser.parse_args()

    # end of arg handling

    if args.parity_evm:
        vm = VMUtils.ParityVM(args.parity_evm, not args.no_docker)
    else:
        vm = VMUtils.GethVM(args.geth_evm, not args.no_docker)

    api = utils.getApi(args.web3)

    if args.test:
        artefacts = test(vm, api)
        import pprint
        pprint.PrettyPrinter().pprint(artefacts)
        sys.exit(0)

    if app and args.www:
        if ':' in args.www:
            host, port = args.www.split(':')
            port = port
        else:
            host = args.www
            port = 5000

        app.debug = args.debug
        app.api = api
        app.vm = vm
        app.run(host=host, port=port)

    elif args.hash:
        artefacts, vm_args = reproduce.reproduceTx(args.hash, vm, api)
        saved_files = utils.saveFiles(OUTPUT_DIR, artefacts)

        # Some tricks to get the right command for local replay
        p_gen = saved_files['parity genesis']['name']
        g_gen = saved_files['geth genesis']['name']
        vm_args['genesis'] = "%s/%s" % (OUTPUT_DIR, g_gen)

        print("\nCommand to execute locally (geth):\n")
        print("%s" % " ".join(vm.makeCommand(**vm_args)))
        print("\nWith memory:\n")
        vm_args['memory'] = True
        print("%s" % " ".join(vm.makeCommand(**vm_args)))
        vm_args.pop('json', None)
        vm_args.pop('memory', None)
        vm_args['statdump'] = "true"
        print("\nFor benchmarking:\n")
        print("%s" % " ".join(vm.makeCommand(**vm_args)))

        print("\nFor opviewing:\n")
        print("python3 opviewer.py -f %s/%s" % (saved_files['json-trace']['path'], saved_files['json-trace']['name']))

        print("\nFor opviewing with sources:\n")
        print(
            "python3 opviewer.py -f %s/%s --web3 '%s' -s path_to_contract_dir -j path_to_solc_combined_json --hash %s" % (
            saved_files['json-trace']['path'], saved_files['json-trace']['name'], args.web3, args.hash))

        logger.debug("creating zip archive for artefacts")
        prefix = args.hash[:8]
        output_archive = os.path.join(OUTPUT_DIR, "%s.zip" % prefix)
        # create a list of files to pack with zipFiles
        input_files = [(os.path.join(v['path'], v['name']), v['name']) for v in saved_files]
        create_zip_archive(input_files=input_files, output_archive=output_archive)

        print("\nZipped files into %s" % output_archive)

    else:
        parser.print_usage()


if __name__ == '__main__':
    logging.basicConfig(format='[%(filename)s - %(funcName)20s() ][%(levelname)8s] %(message)s',
                        level=logging.INFO)
    main()



#!/usr/bin/env python3
"""
This is a tool to replicate live on-chain events. It starts with a transaction id

1 Fetch the transaction data from an API. 
2 Fetch the data at destination
3 Fetch balance and nonce at source
4 Execute transaction
5 If transaction has any externally reaching ops (BALANCE, EXTCODECOPY, CALL etc), 
  * Fetch info about those accounts
6. Go back to 4 until all account info is fetched. Code, balance, nonce etc. 
7. Generate a genesis, and code to invoke the on-chain event.

"""

import json
import tempfile, os, traceback

from sys import argv, exit

from . import compiler as c
from . import genesis as gen
from . import opcodes
from . import evmtrace
#from . import multiapi
from . import utils

def findExternalCalls(list_of_output):
    externals = {
                "CALL"         : lambda o : o['stack'][-2], 
                "CALLCODE"     : lambda o : o['stack'][-2], 
                "DELEGATECALL" : lambda o : o['stack'][-2], 
                "STATICCALL"   : lambda o : o['stack'][-2], 
                "EXTCODECOPY"  : lambda o : o['stack'][-1],
                "EXTCODESIZE"  : lambda o : o['stack'][-1],
                "BALANCE"      : lambda o : o['stack'][-1],
                }
    accounts = set()
    for l in list_of_output:
        if len(l) == 0 or l[0] != "{":
            print("Odd line: %s" % l)
            continue

        o = json.loads(l.strip())

        if 'opName' in o and o['opName'] in externals.keys():
            accounts.add(externals[o['opName']](o))

    accounts.discard('0x0')
    return accounts

def findStorageLookups(list_of_output, original_context):
    """ This method searches through an EVM-output and locates SLOAD queries
    Returns a list of (<address>, <key>)
    """
    # call context
    context = [] 
    # duplicate avoidance
    refs = set()

    # In order to do this, we need to track the context (address) we're executing in
    callstack = []

    cur_depth = 1
    prev_depth = 1
    prev_op = None
    cur_address = original_context

    for l in list_of_output:
        if len(l) == 0 or l[0] != "{":
            continue

        o = json.loads(l.strip())
        if not 'depth' in o.keys():
            #We're done here
            break
        # Address tracking - what's the context we're executing in
        cur_depth = o['depth']
        if cur_depth > prev_depth:
            #Made it into a call-variant
            callstack.append(cur_address)
            # All call-lookalikes are 'gas,address,value' on stack, 
            # so address is second item of prev line
            #
            # There's two exceptions, though; CREATE and DELEGATECALL
            # With a CREATE, we don't know the address until after the RETURN
            # and it isn't necessary to load storage b/c it hasn't been initialized
            # After a DELEGATECALL, we've increased the depth, 
            # but still operating on the same context, regardless of the
            # address that was invoked
            if prev_op['op'] == 0xf0:
                cur_address = None
            elif prev_op['op'] != 0xf4:
                cur_address = prev_op['stack'][-2]

        if cur_depth < prev_depth:
            # Returned from a call
            cur_address = callstack.pop()

        # Sload tracking
        if cur_address and (o['opName'] in ['SLOAD','SSTORE'] or o['op'] in [0x54, 0x55]):

            key  = o['stack'][-1]
            entry = (cur_address, key)
            refs.add(entry)

        prev_op = o
        prev_depth = cur_depth

    return refs


def debugdump(obj):
    import pprint
    pprint.PrettyPrinter().pprint(obj)


def reproduceTx(txhash, vm, api):

    genesis = gen.Genesis()
    
    tx = api.getTransaction(txhash)

    s = tx['from']
    r = tx['to']
    tx['input'] = tx['input'][2:]

    if r == '0x0':
        r = None

    create = r is None

    #debugdump(tx)
    blnum = int(tx['blockNumber'])

    if blnum > 4370000:
        genesis.setConfigMetropolis()
        print("Setting Metro config for block", blnum)
    elif blnum > 2675000:
        genesis.setConfigSpuriousDragon()
        print("Setting Spurious config for block", blnum)
    elif blnum > 2463000:
        genesis.setConfigTangerineWhistle()
        print("Setting Tangerine config for block", blnum)
    elif blnum > 1150000: 
        print("Setting Homestead config for block", blnum)



    externals_fetched = set()
    externals_tofetch = set([s,r])
    externals_tofetch.discard(None)

    storage_slots_fetched = set()
    slots_to_fetch = set()
    receivercode = ""
    done = False
    while not done:
        done = True
        # Add accounts that we know of 
        for addr in list(externals_tofetch):
            acc = api.getAccountInfo( addr , blnum - 1)  # need to load accountInfo at block before tx
            genesis.add(acc)
            #debugdump(acc)
            done = False

        
        externals_fetched.update(externals_tofetch)

        for (addr,key) in list(slots_to_fetch):
            val = api.getStorageSlot( addr, int(key,16), blnum - 1)  # need to load storage at block before tx
            genesis.addStorage(addr, key, val)
            done = False
        storage_slots_fetched.update(slots_to_fetch)
        
        (g_path, p_path) = genesis.export(txhash[:8])
        genesis_path = g_path
        if vm.genesis_format == 'parity':
            genesis_path = p_path

        vm_args = {
            "receiver"  : r,
            "sender"    : s,
            "genesis"   : genesis_path, 
            "json"      : True, 
            "input"     : tx['input'],
            "gas"       : tx['gas'], 
            "memory"    : False,
            "create"    : create,
        }
        if not done:
            print("Executing tx...")
        else:
            #One final trace with memory on, makes for better annotated trace
            print("Final execution (memory on)")
            vm_args['memory'] = True

        # We could use the following to set the code for parity:
        #receivercode = genesis.codeAt(r)
        #print(tx)
        output =  vm.execute(**vm_args)

        fd, temp_path = tempfile.mkstemp( prefix=txhash[:8]+'_', suffix=".txt")
        with open(temp_path, 'w') as f :
            f.write("\n".join(output))
            print("Saved trace to %s" % temp_path)
        os.close(fd)

        if not done:
            # External accounts to lookup
            externals_found = findExternalCalls(output)
            externals_tofetch = externals_found.difference(externals_fetched)
            if len(externals_tofetch) > 0:
                print("External accounts to fetch: %s " % externals_tofetch )

            # Storage slots to lookup
            slots_found = findStorageLookups(output, r)
            slots_to_fetch = slots_found.difference(storage_slots_fetched)
            if len(slots_to_fetch) > 0:
                print("SLOTS to fetch: %s " % slots_to_fetch)


    artefacts = {
        'geth genesis'   : g_path, 
        'parity genesis' : p_path, 
        'json-trace': temp_path}

    try:
        annotated_trace = evmtrace.traceEvmOutput(temp_path)
        fd, a_trace = tempfile.mkstemp( prefix=txhash[:8]+'_', suffix=".evmtrace.txt")
        with open(a_trace, 'w') as f :
            f.write(str(annotated_trace))
        os.close(fd)
        artefacts['annotated trace'] = a_trace

    except Exception as e:
        print("Evmtracing failed")
        traceback.print_exc()

    print("vm args:")
    print(vm_args)
    return artefacts, vm_args

def testStoreLookup():
    tr = "/data/workspace/evmlab/0xd6d519043d40691a36c9e718e47110309590e6f47084ac0ec00b53718e449fd3_der80goh.txt"
    with open(tr, "r") as f:
        l = findStorageLookups(f,"recipient")
        debugdump(l)


def fetch(args):
    if len(args) < 1:
        print("Usage: ./reproduce.py <tx hash>")
        exit(1)
    evmbin = "evm"
    tx = args[0]
    api = utils.getApi("https://mainnet.infura.io:443")
    reproduceTx(tx, evmbin, api)


if __name__ == '__main__':
    fetch(argv[1:])
    test()
    #testStoreLookup()



STOP          = 0x00 # Pops: 0, Pushes: 0, Gas: 0],
ADD           = 0x01 # Pops: 2, Pushes: 1, Gas: 3],
MUL           = 0x02 # Pops: 2, Pushes: 1, Gas: 5],
SUB           = 0x03 # Pops: 2, Pushes: 1, Gas: 3],
DIV           = 0x04 # Pops: 2, Pushes: 1, Gas: 5],
SDIV          = 0x05 # Pops: 2, Pushes: 1, Gas: 5],
MOD           = 0x06 # Pops: 2, Pushes: 1, Gas: 5],
SMOD          = 0x07 # Pops: 2, Pushes: 1, Gas: 5],
ADDMOD        = 0x08 # Pops: 3, Pushes: 1, Gas: 8],
MULMOD        = 0x09 # Pops: 3, Pushes: 1, Gas: 8],
EXP           = 0x0a # Pops: 2, Pushes: 1, Gas: 10],
SIGNEXTEND    = 0x0b # Pops: 2, Pushes: 1, Gas: 5],
LT            = 0x10 # Pops: 2, Pushes: 1, Gas: 3],
GT            = 0x11 # Pops: 2, Pushes: 1, Gas: 3],
SLT           = 0x12 # Pops: 2, Pushes: 1, Gas: 3],
SGT           = 0x13 # Pops: 2, Pushes: 1, Gas: 3],
EQ            = 0x14 # Pops: 2, Pushes: 1, Gas: 3],
ISZERO        = 0x15 # Pops: 1, Pushes: 1, Gas: 3],
AND           = 0x16 # Pops: 2, Pushes: 1, Gas: 3],
OR            = 0x17 # Pops: 2, Pushes: 1, Gas: 3],
XOR           = 0x18 # Pops: 2, Pushes: 1, Gas: 3],
NOT           = 0x19 # Pops: 1, Pushes: 1, Gas: 3],
BYTE          = 0x1a # Pops: 2, Pushes: 1, Gas: 3],
SHA3          = 0x20 # Pops: 2, Pushes: 1, Gas: 30],
ADDRESS       = 0x30 # Pops: 0, Pushes: 1, Gas: 2],
BALANCE       = 0x31 # Pops: 1, Pushes: 1, Gas: 20],
ORIGIN        = 0x32 # Pops: 0, Pushes: 1, Gas: 2],
CALLER        = 0x33 # Pops: 0, Pushes: 1, Gas: 2],
CALLVALUE     = 0x34 # Pops: 0, Pushes: 1, Gas: 2],
CALLDATALOAD  = 0x35 # Pops: 1, Pushes: 1, Gas: 3],
CALLDATASIZE  = 0x36 # Pops: 0, Pushes: 1, Gas: 2],
CALLDATACOPY  = 0x37 # Pops: 3, Pushes: 0, Gas: 3],
CODESIZE      = 0x38 # Pops: 0, Pushes: 1, Gas: 2],
CODECOPY      = 0x39 # Pops: 3, Pushes: 0, Gas: 3],
GASPRICE      = 0x3a # Pops: 0, Pushes: 1, Gas: 2],
EXTCODESIZE   = 0x3b # Pops: 1, Pushes: 1, Gas: 20],
EXTCODECOPY   = 0x3c # Pops: 4, Pushes: 0, Gas: 20],
RETURNDATASIZE= 0x3D
RETURNDATACOPY= 0x3E
BLOCKHASH     = 0x40 # Pops: 1, Pushes: 1, Gas: 20],

COINBASE      = 0x41 # Pops: 0, Pushes: 1, Gas: 2],
TIMESTAMP     = 0x42 # Pops: 0, Pushes: 1, Gas: 2],
NUMBER        = 0x43 # Pops: 0, Pushes: 1, Gas: 2],
DIFFICULTY    = 0x44 # Pops: 0, Pushes: 1, Gas: 2],
GASLIMIT      = 0x45 # Pops: 0, Pushes: 1, Gas: 2],
POP           = 0x50 # Pops: 1, Pushes: 0, Gas: 2],
MLOAD         = 0x51 # Pops: 1, Pushes: 1, Gas: 3],
MSTORE        = 0x52 # Pops: 2, Pushes: 0, Gas: 3],
MSTORE8       = 0x53 # Pops: 2, Pushes: 0, Gas: 3],
SLOAD         = 0x54 # Pops: 1, Pushes: 1, Gas: 50],
SSTORE        = 0x55 # Pops: 2, Pushes: 0, Gas: 0],
JUMP          = 0x56 # Pops: 1, Pushes: 0, Gas: 8],
JUMPI         = 0x57 # Pops: 2, Pushes: 0, Gas: 10],
PC            = 0x58 # Pops: 0, Pushes: 1, Gas: 2],
MSIZE         = 0x59 # Pops: 0, Pushes: 1, Gas: 2],
GAS           = 0x5a # Pops: 0, Pushes: 1, Gas: 2],
JUMPDEST      = 0x5b # Pops: 0, Pushes: 0, Gas: 1],
LOG0          = 0xa0 # Pops: 2, Pushes: 0, Gas: 375],
LOG1          = 0xa1 # Pops: 3, Pushes: 0, Gas: 750],
LOG2          = 0xa2 # Pops: 4, Pushes: 0, Gas: 1125],
LOG3          = 0xa3 # Pops: 5, Pushes: 0, Gas: 1500],
LOG4          = 0xa4 # Pops: 6, Pushes: 0, Gas: 1875],
CREATE        = 0xf0 # Pops: 3, Pushes: 1, Gas: 32000],
CREATE2       = 0xfb # 
CALL          = 0xf1 # Pops: 7, Pushes: 1, Gas: 40],
CALLCODE      = 0xf2 # Pops: 7, Pushes: 1, Gas: 40],
RETURN        = 0xf3 # Pops: 2, Pushes: 0, Gas: 0],
DELEGATECALL  = 0xf4 # Pops: 6, Pushes: 0, Gas: 40],
STATICCALL    = 0xfa
REVERT        = 0xfd
SUICIDE       = 0xff # Pops: 1, Pushes: 0, Gas: 0],
SELFDESTRUCT  = 0xff # Pops: 1, Pushes: 0, Gas: 0],

PUSH1         =0x60
PUSH32        =0x7f

DUP1  = 0x80
DUP2  = 0x81
DUP3  = 0x82
DUP4  = 0x83
DUP5  = 0x84
DUP6  = 0x85
DUP7  = 0x86
DUP8  = 0x87
DUP9  = 0x88
DUP10 = 0x89
DUP11 = 0x8a
DUP12 = 0x8b
DUP13 = 0x8c
DUP14 = 0x8d
DUP15 = 0x8e
DUP16 = 0x8f

SWAP1  = 0x90
SWAP2  = 0x91
SWAP3  = 0x92
SWAP4  = 0x93
SWAP5  = 0x94
SWAP6  = 0x95
SWAP7  = 0x96
SWAP8  = 0x97
SWAP9  = 0x98
SWAP10 = 0x99
SWAP11 = 0x9a
SWAP12 = 0x9b
SWAP13 = 0x9c
SWAP14 = 0x9d
SWAP15 = 0x9e
SWAP16 = 0x9f

import sys

def bytecode(value):

	typ = type(value)
	if typ == str and value[:2] == "0x":
		value = value[2:]

	if typ in [float, int]:
		value = '{0:02x}'.format(int(value))

	if sys.version_info < (3, 0):
		if typ == unicode and value[:2] == "0x":
			value = value[2:]
		if typ == long:
			value = '{0:02x}'.format(int(value))

	value = ('0' * (len(value) % 2)) + value
	return value

class Program():

	def __init__(self):
		self.compiled = []
		self.ops = []
		self.mstore= lambda index,value: self.push(value).push(index).op(MSTORE)
		self.mstore8= lambda index,value: self.push(value).push(index).op(MSTORE8)
		self.add =   lambda x,y: self.push(y).push(x).op(ADD)
		self.sub =   lambda x,y: self.push(y).push(x).op(SUB)
		self.mul =   lambda x,y: self.push(y).push(x).op(MUL)
		self.div =   lambda x,y: self.push(y).push(x).op(DIV)
		self.sdiv =  lambda x,y: self.push(y).push(x).op(SDIV)
		self.mod =   lambda x,y: self.push(y).push(x).op(MOD)
		self.smod =  lambda x,y: self.push(y).push(x).op(SMOD)
		self.exp =   lambda x,y: self.push(y).push(x).op(EXP)

		self.create       = lambda v,p,s:self.push(s).push(p).push(v).op(CREATE)
		self.codecopy     = lambda t,f,s:self.push(s).push(f).push(t).op(CODECOPY)
		self.extcodecopy  = lambda a,t,f,s:self.push(s).push(f).push(t).push(a).op(CODECOPY)
		self.extcoodesize = lambda a:self.push(a).op(EXTCODESIZE)
		self.selfdestruct = lambda a: self.push(a).op(SUICIDE)

		#log without topics and data mem[p..(p+s))
		self.log0        = lambda p, s: self.push(s).push(p).op(LOG0)                  
		#log with topic t1 and data mem[p..(p+s))
		self.log1        = lambda p, s, t1: self.push(t1).push(s).push(p).op(LOG1)
		#log with topics t1, t2 and data mem[p..(p+s))
		self.log2        = lambda p, s, t1, t2: self.push(t2).push(t1).push(s).push(p).op(LOG3)
		#log with topics t1, t2, t3 and data mem[p..(p+s))
		self.log3        = lambda p, s, t1, t2, t3:  self.push(t3).push(t2).push(t1).push(s).push(p).op(LOG3)
		#log with topics t1, t2, t3, t4 and data mem[p..(p+s))
		self.log4        = lambda p, s, t1, t2, t3, t4: self.push(t4).push(t3).push(t2).push(t1).push(s).push(p).op(LOG3)
		self.jump        = lambda label : self.push(label).op(JUMP)
		self.jumpi        = lambda label,cond : self.push(cond).push(label).op(JUMPI)
		self.revert      = lambda  memStart, memSize: self.push(memSize).push(memStart).op(REVERT)

		
	def _add(self, x):
		if x == None:
			return self

		if type(x) == str:
			self.compiled.append(x)
		else:
			self.compiled.append(bytecode(x))

		return self

	def extend(self,program):
		self.compiled.extend(program.compiled)

	def _addOp(self,op,v = None):
		self._add(op)
		self._add(v)
		return self

	def op(self,x):
		self._add(x)
		return self

	def push(self,value):
		value = bytecode(value)
		length = len(value) / 2

		assert length <=32

		self._addOp(PUSH1+(length-1), bytecode(value));
		return self



	def call(self,gas ,address,value = 0,instart = 0, insize = 0, out = 0, outsize = 0):
		self.push(outsize)
		self.push(out)
		self.push(insize)
		self.push(instart)
		self.push(value)
		self.push(address)
		if gas is not None:
			self.push(gas)
		else:
			self.op(GAS)
		self._addOp(CALL)
		return self

	def callcode(self,gas ,address,value = 0,instart = 0, insize = 0, out = 0, outsize = 0):
		self.push(outsize)
		self.push(out)
		self.push(insize)
		self.push(instart)
		self.push(value)
		self.push(address)
		self.push(gas)
		self._addOp(CALLCODE)
		return self

	def delegatecall(self,gas ,address,instart = 0, insize = 0, out = 0, outsize = 0):
		self.push(outsize)
		self.push(out)
		self.push(insize)
		self.push(instart)
		self.push(address)
		if gas is not None:
			self.push(gas)
		else:
			self.op(GAS)
		self._addOp(DELEGATECALL)
		return self

	def staticcall(self,gas ,address,instart = 0, insize = 0, out = 0, outsize = 0):
		self.push(outsize)
		self.push(out)
		self.push(insize)
		self.push(instart)
		self.push(address)
		if gas is not None:
			self.push(gas)
		else:
			self.op(GAS)
		self._addOp(STATICCALL)
		return self

	def rreturn(self, memStart=0, memSize=0):
		self.push(memSize)
		self.push(memStart)
		self._addOp(RETURN)
		return self


	def bytecode(self):
		return "".join(self.compiled)

	def label(self):
		return len(self.bytecode()) / 2

	def jumpdest(self):
		here = self.label()
		self.op(JUMPDEST)
		return here
	

	def __str__(self):
		return ",".join(self.compiled)





import os
import json
from .opcodes import opcodes
from . import compiler

OPCODE_FORMATS = {
    "ADD":          "{0} + {1}",
    "SUB":          "{0} - {1}",
    "EXP":          "{0} ** {1}",
    "MUL":          "{0} * {1}",
    "DIV":          "{0} / {1}",
    "EQ":           "{0} == {1}",
    "LT":           "{0} < {1}",
    "GT":           "{0} > {1}",
    "AND":          "{0} & {1}",
    "OR":           "{0} | {1}",
    "XOR":          "{0} ^ {1}",
    "NOT":          "!{0}",
    "ISZERO":       "{0} == 0",
    "JUMP":         "JUMP(pc={0})",
    "JUMPI":        "JUMPI(pc={0}, cond={1})",
    "MLOAD":        "MLOAD(off={0})",
    "MSTORE":       "MSTORE(off={0}, data={1})",
    "SHA3":         "SHA3(off={0}, len={1})",
    "SLOAD":        "SLOAD(addr={0})",
    "SSTORE":       "SSTORE(addr={0}, data={1})",
    "CALL":         "CALL(gas={0}, addr={1}, value={2}, in={3}, insize={4}, out={5}, outsize={6})",
    "DELEGATECALL": "DELEGATECALL(gas={0}, addr={1}, in={2}, insize={3}, out={4}, outsize={5})",
    "STATICCALL":   "STATICCALL(gas={0}, addr={1}, in={2}, insize={3}, out={4}, outsize={5})",
    "CALL":         "CALL(gas={0}, addr={1}, value={2}, in={3}, insize={4}, out={5}, outsize={6})",
    "EXTCODECOPY":  "EXTCODECOPY(addr={0}, to={1}, from={2}, size={3}))",
    "CODECOPY":     "CODECOPY(to={1}, from={2}, size={3}))",
    "CREATE":       "CREATE(val={0}, offset={1}, size={2})"
}

def opinfo(opcode):
    if opcode in opcodes.keys():
        return opcodes[opcode]
    return "INVALID",0,0,0

class Annotable(object):
    def __init__(self):
        self.annotations = {}
        super(Annotable, self).__init__()

    def setAnnotation(self, obj):
        self.annotations[type(obj)] = obj


class OpcodeNode(Annotable):
    def __init__(self, pc, depth, opcode, args, result):
        super(OpcodeNode, self).__init__()
        self.pc = pc
        self.opcode = opcode
        self.depth = depth
        self.opname, self.ins, self.outs, self.gas = opinfo(opcode)
        self.args = args
        self.result = result

    def __str__(self):
        if self.opname in OPCODE_FORMATS:
            fmt = OPCODE_FORMATS[self.opname].format(*self.args)
        else:
            fmt = "{0}({1})".format(self.opname, ', '.join(self.args))
        if self.result:
            return "{0:40} -> {1}".format(fmt, ', '.join(self.result))
        else:
            return fmt

    def toHtml(self):
        if self.opname in OPCODE_FORMATS:
            fmt = OPCODE_FORMATS[self.opname].format(*self.args)
        else:
            fmt = "{0}({1})".format(self.opname, ', '.join(self.args))
        if self.result:
            return "{0:40} -> {1}".format(fmt, ', '.join(self.result))
        else:
            return fmt

class CallNode(OpcodeNode):
    def __init__(self, pc, depth, opcode, args, result, ops):
        super(CallNode, self).__init__(pc, depth, opcode, args, result)
        self.ops = ops


class PushNode(OpcodeNode):
    def __init__(self, pc, depth, opcode, args, result):
        super(PushNode, self).__init__(pc, depth, opcode, args, result)

    def __str__(self):
        return self.result[0]

    def toHtml(self):
        return str(self)


def buildAST(trace):
    ops = []
    stack = []
    pc = 0

    for step in trace:
        pc = step.get('pc', pc)
        if step['op'] in opcodes.keys():
            opname, ins, outs, gas = opcodes[step['op']]
        else:
            opname, ins, outs,gas = "INVALID",0,0,0

        if ins > 0:
            args = stack[-ins:][::-1]
            stack = stack[:-ins]
        else:
            args = []
        if 'ops' in step:
            ops.append(CallNode(pc, step['depth'], step['op'], args, step['result'], buildAST(step['ops'])))
        elif opname.startswith('PUSH'):
            ops.append(PushNode(pc, step['depth'], step['op'], args, step['result']))
        else:
            ops.append(OpcodeNode(pc, step['depth'], step['op'], args, step['result']))
        pc += step.get('len', 1)
        stack.extend(step['result'][::-1])
    return ops


class TransactionTrace(Annotable):
    def __init__(self, ops):
        super(TransactionTrace, self).__init__()
        self.ops = ops

    @classmethod
    def build(cls, trace):
        return TransactionTrace(buildAST(trace))

    def __str__(self):
        lines = []
        stack = [(self, 0)]
        while stack:
            call, startidx = stack.pop()
            for i in range(startidx, len(call.ops)):
                op = call.ops[i]
                lines.append("{0:>4} 0x{1:0>4x} {2}{3}".format(op.depth,op.pc, '  '*len(stack), str(op)))
                if hasattr(op, 'ops'):
                    stack.append((call, i + 1))
                    stack.append((op, 0))
                    break
                elif hasattr(op, 'expression') and hasattr(op.expression, 'ops'):
                    stack.append((call, i + 1))
                    stack.append((op.expression, 0))
                    break
        return '\n'.join(lines)

    def iterator(self):
        #lines = []
        stack = [(self, 0)]
        while stack:
            call, startidx = stack.pop()
            for i in range(startidx, len(call.ops)):
                op = call.ops[i]
                yield (len(stack), op)
                #lines.append("0x{0:0>4x} {1}{2}".format(op.pc, '  '*len(stack), str(op)))
                if hasattr(op, 'ops'):
                    stack.append((call, i + 1))
                    stack.append((op, 0))
                    break
                elif hasattr(op, 'expression') and hasattr(op.expression, 'ops'):
                    stack.append((call, i + 1))
                    stack.append((op.expression, 0))
                    break
        #return '\n'.join(lines)


class ReachingDefinitions(list):
    """Encapsulates a list of the sources of each argument to an operation."""


class ReachesDefinitions(list):
    """Encapsulates a list of the consumers of an operation's output."""


class VariableName(str):
    """Annotation for variable name assignments."""


class AssignmentStatement(object):
    def __init__(self, depth, pc, varname, expression):
        self.pc = pc
        self.depth = depth
        self.varname = varname
        self.expression = expression

    def __str__(self):
        return "{0} = {1}".format(self.varname, self.expression)


class ExpressionStatement(object):
    def __init__(self, depth, pc, expression):
        self.pc = pc
        self.depth = depth
        self.expression = expression

    def __str__(self):
        return str(self.expression)


class VariableExpression(object):
    def __init__(self, depth, varname):
        self.varname = varname
        self.depth = depth

    def __str__(self):
        return self.varname


class OperationExpression(object):
    def __init__(self, depth, op, args):
        self.op = op
        self.depth = depth
        self.args = args

    @property
    def opname(self):
        return self.op.opname

    @property
    def pc(self):
        return self.op.pc

    def __str__(self):
        if self.op.opname in OPCODE_FORMATS:
            return OPCODE_FORMATS[self.op.opname].format(*self.args)
        else:
            return "{0}({1})".format(self.opname, ', '.join(map(str, self.args)))


class CallExpression(OperationExpression):
    def __init__(self, depth,op, args, ops):
        super(CallExpression, self).__init__(depth, op, args)
        self.ops = ops


class LiteralExpression(OperationExpression):
    def __init__(self, depth, value):
        self.value = value
        self.depth = depth

    def __str__(self):
        return self.value


def findReachings(ast):
    stack = []
    for op in ast.ops:
        if op.ins > 0:
            args = stack[-op.ins:]
            stack = stack[:-op.ins]
        else:
            args = []
        if op.opname.startswith('DUP'):
            stack.extend(args)
            stack.append(args[0])
        elif op.opname.startswith('SWAP'):
            stack.append(args[-1])
            stack.extend(args[1:-1])
            stack.append(args[0])
        else:
            for arg in args:
                arg.annotations[ReachesDefinitions].append(op)
            op.setAnnotation(ReachingDefinitions(args[::-1]))
            op.setAnnotation(ReachesDefinitions())
            stack.extend([op] * op.outs)
        if hasattr(op, 'ops'):
            findReachings(op)


def nameIterator():
    prefixIterator = nameIterator()
    prefix = ""
    while True:
        for i in "abcdefghijklmnopqrstuvwxyz":
            yield prefix + i
        prefix = next(prefixIterator)


def buildExpression(op):
    subexps = []
    for arg in op.annotations[ReachingDefinitions]:
        if VariableName in arg.annotations:
            subexps.append(VariableExpression(op.depth, arg.annotations[VariableName]))
        else:
            subexps.append(buildExpression(arg))
    if isinstance(op, CallNode):
        return CallExpression(op.depth, op, subexps, composeOperations(op.ops))
    elif isinstance(op, PushNode):
        return LiteralExpression(op.depth, op.result[0])
    else:
        return OperationExpression(op.depth, op, subexps)


def composeOperations(ops):
    varnames = nameIterator()
    statements = []
    for op in ops:
        # Ignore SWAP and DUP, which don't have annotations
        if ReachingDefinitions not in op.annotations:
            continue
        reaches = op.annotations[ReachesDefinitions]
        if not op.opname.startswith('PUSH'):
            if len(reaches) == 0:
                statements.append(ExpressionStatement(op.depth, op.pc, buildExpression(op)))
            elif len(reaches) > 1 or isinstance(op, CallNode):
                varname = next(varnames)
                op.setAnnotation(VariableName(varname))
                statements.append(AssignmentStatement(op.depth, op.pc, varname, buildExpression(op)))
    return statements


def trace(web3, txid, compose = True):
    result = web3._requestManager.request_blocking('debug_traceTransaction', (txid, {'tracer': tracer}))
    ast = TransactionTrace.build(result)
    findReachings(ast)
    if compose: 
        ast = TransactionTrace(composeOperations(ast.ops))
 
    return ast

def traceEvmOutput(tracefile, compose = True):
    result = evmResult(tracefile)
    ast = TransactionTrace.build(result)
    findReachings(ast)
    if compose: 
        ast = TransactionTrace(composeOperations(ast.ops))
 
    return ast

def evmResult(tracefile):

    def isPush(op):
        return op >= compiler.PUSH1 and op <= compiler.PUSH32 

    res = {
        'stack' : [{"ops" : []}]
    }
    npushes = {0: 0, 1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1, 7: 1, 8: 1, 9: 1, 10: 1, 11: 1, 16: 1, 17: 1, 18: 1, 19: 1, 20: 1, 21: 1, 22: 1, 23: 1, 24: 1, 25: 1, 26: 1, 32: 1, 48: 1, 49: 1, 50: 1, 51: 1, 52: 1, 53: 1, 54: 1, 55: 0, 56: 1, 57: 0, 58: 1, 59: 1, 60: 0, 61: 1, 62: 0, 64: 1, 65: 1, 66: 1, 67: 1, 68: 1, 69: 1, 80: 0, 81: 1, 82: 0, 83: 0, 84: 1, 85: 0, 86: 0, 87: 0, 88: 1, 89: 1, 90: 1, 91: 0, 96: 1, 97: 1, 98: 1, 99: 1, 100: 1, 101: 1, 102: 1, 103: 1, 104: 1, 105: 1, 106: 1, 107: 1, 108: 1, 109: 1, 110: 1, 111: 1, 112: 1, 113: 1, 114: 1, 115: 1, 116: 1, 117: 1, 118: 1, 119: 1, 120: 1, 121: 1, 122: 1, 123: 1, 124: 1, 125: 1, 126: 1, 127: 1, 128: 2, 129: 3, 130: 4, 131: 5, 132: 6, 133: 7, 134: 8, 135: 9, 136: 10, 137: 11, 138: 12, 139: 13, 140: 14, 141: 15, 142: 16, 143: 17, 144: 2, 145: 3, 146: 4, 147: 5, 148: 6, 149: 7, 150: 8, 151: 9, 152: 10, 153: 11, 154: 12, 155: 13, 156: 14, 157: 15, 158: 16, 159: 17, 160: 0, 161: 0, 162: 0, 163: 0, 164: 0, 240: 1, 241: 1, 242: 1, 243: 0, 244: 0, 255: 0}

    with open(tracefile) as f:
        for line in f:
        
            log = json.loads(line)
            if 'output' in log.keys():
                return res['stack'][0]['ops']

            def peek(n):
                return int(log['stack'][-1-n],16)

            def isOp(op):
                return log['op'] == op

            d = ' '.ljust(log['depth']*4)

            if log['depth'] != len(res['stack']):
                res['stack'] = res['stack'][0:-1]
            #print("line", line)
            try:
                frame = res['stack'][-1]
            except Exception as e:
                print(res)
                print(line)
                raise e

            if log['depth'] == len(res['stack']):
                opinfo = {
                "op" : log['op'], 
                "depth" : log['depth'],
                'result' : [],
                }
                if len(frame['ops']) > 0:
                    prevop = frame['ops'][-1]
                    for i in range(0,npushes[prevop['op']]):
                        prevop['result'].append(hex(peek(i)))

                if isOp(compiler.CALL) or isOp(compiler.CALLCODE) or isOp(compiler.DELEGATECALL) or isOp(compiler.STATICCALL):
                    opinfo["error"] = None;
                    opinfo["return"] = None;
                    opinfo["ops"] = [];
                    opinfo["gas"]   = peek(0)
                    opinfo["to"]    = peek(1)

                    if isOp(compiler.DELEGATECALL) or isOp(compiler.STATICCALL):
                        instart = peek(-2)
                        insize  = peek(-3)
                    else:
                        opinfo["value"] = peek(2)
                        instart = peek(3)
                        insize  = peek(4)

                    opinfo["input"] = log['memory'][instart: instart + insize];
                    res['stack'].append(opinfo)

                elif isOp(compiler.RETURN):
                    out = peek(0)
                    outsize = peek(1)
                    print("Out" , out)
                    print("Outsize", outsize)
                    frame['return'] = log['memory'][out:out+outsize]
                elif isOp(compiler.STOP) or isOp(compiler.SELFDESTRUCT):
                    frame['return'] = None
                elif isOp(compiler.JUMPDEST):
                    opinfo['pc'] = log['pc']

                if isPush(log['op']):
                    opinfo['len'] = log['op'] - 0x5e

                frame['ops'].append(opinfo)

        
def testFile(fname):
    testfile = os.path.join(os.path.dirname(__file__), fname)
    ast = traceEvmOutput(testfile)
    print(str(ast))
    print("OK")

if __name__ == '__main__':
#    testFile("example_trace.txt")
    testFile("example_trace2.txt")



#!/usr/bin/env python
"""
deps:
*) pip install --upgrade pip wheel setuptools twine
publish:
1) #> python setup.py sdist bdist_wheel
2) #> twine upload dist/*   #<specific specific wheel if needed>; --repository <testpypi> or  --repository-url <testpypi-url>
"""
import os
from setuptools import setup


def read(fname):
    fpath = os.path.join(os.path.dirname(__file__), fname)
    fd = open(fpath, 'r')
    data = fd.read()
    fd.close()
    return data


version = '0.3.0.0.1'  # keep this in sync with the last release tag!

setup(name='Evmlab',
      version=version,
      description='Ethereum EVM utilities',
      author='Martin Holst Swende',
      author_email='martin.swende@ethereum.org',
      license="GPLv3",
      keywords=["ethereum", "transaction", "debugger"],
      url="https://github.com/ethereum/evmlab/",
      download_url="https://github.com/ethereum/evmlab/tarball/v%s" % version,
      # generate rst from .md:  pandoc --from=markdown --to=rst README.md -o README.rst (fix diff section and footer)
      long_description=read("README.md") if os.path.isfile("README.md") else "",
      # for pypi.org; empty because we do not ship Readme.md with the package. may otherwise fail on install
      long_description_content_type='text/markdown',  # requires twine and recent setuptools
      packages=['evmlab',
                'evmlab.tools',
                'evmlab.tools.reproducer'],
      package_data={'evmlab.tools.reproducer': ["templates/*"]},
      install_requires=["requests", "web3", "eth-hash[pycryptodome]", "rlp>=1.0"],
      extras_require={"consolegui": ["urwid"],
                      "abidecoder": ["ethereum-input-decoder"],
                      "docker": ["docker==3.0.0"]}
      )



from functools import reduce
import json


def int_or_none(v):
    if v in ('i', 'o', '-'):
        return v
    if v == '':
        return None
    return int(v)

def add_missing_values(arr):
    missing = 4 - len(arr)
    return arr + [None] * missing

def is_push(inst):
    return inst >= 0x60 and inst < 0x7f

def pushdata_length(inst):
    return inst - 0x5f

def instruction_length(inst):
    if is_push(inst):
        return 1 + pushdata_length(inst)
    return 1


class SourceMap(object):

    def __init__(self, source, data, contract):
        self.source = source
        self.data = data
        self.contract = contract
        self.source_line_offsets = self._compute_source_line_offsets(source)
        self.byte_to_instr = self._compute_byte_to_instr()

    @classmethod
    def from_standard_json(self, source, data, contract):
        return SourceMap(source, data, contract)

    @property
    def bin(self):
        from binascii import unhexlify
        bytes = self.data['contracts'][self.contract]['bin-runtime']
        # bytes = bytes.replace('__test.sol:SafeMath_____________________', '0000000000000000000000000000000000000000')
        return unhexlify(bytes)

    @property
    def srcmap(self):
        _srcmap = self.data['contracts'][self.contract]['srcmap-runtime']
        return self._fill_srcmap(_srcmap)

    def _fill_srcmap(self, srcmap):
        elems = srcmap.split(';')
        elems = [[int_or_none(v) for v in e.split(':')] for e in elems]
        elems = [add_missing_values(e) for e in elems]

        # fill missing values
        prev_elem = elems[0]
        for elem in elems[1:]:
            for i in range(len(elem) - 1):  # don't do anything to last elem ('i', 'o', '-')
                if elem[i] is None:
                    elem[i] = prev_elem[i]
            prev_elem = elem
        return elems

    def _compute_byte_to_instr(self):
        result = []
        byte_idx = 0
        instr_idx = 0

        while byte_idx < len(self.bin):
            length = instruction_length(self.bin[byte_idx])
            for i in range(length):
                result.append(instr_idx)
            byte_idx += length
            instr_idx += 1

        return result

    def _compute_source_line_offsets(self, source):
        offsets = [0]
        for i in range(len(source)):
            if source[i] == '\n':
                offsets.append(i+1)
        return offsets

    def line_number_for_instr(self, instr):
        offset = self.srcmap[instr][0]

        prev_n = self.source_line_offsets[0]
        ln = 1
        for n in self.source_line_offsets[1:]:
            if offset >= prev_n and offset < n:
                return ln
            prev_n = n
            ln += 1
        return ln

    def instr_for_pc(self, pc):
        return self.byte_to_instr[pc]

    def line_for_instr(self, instr):
        offset = self.srcmap[instr][0]

        start = self.source.rfind('\n', 0, offset)
        end = self.source.find('\n', offset)
        return self.source[start+1:end]



from .contract import Contract
from . import mk_contract_address, encode_hex

def buildContexts(ops, api, contracts, txhash):
    contract_stack = []

    tx = api.getTransaction(txhash)
    to = tx['to']
    blnum = int(tx['blockNumber'])

    if to == '0x0':
        to = None

    create = to is None

    # contract deployment
    # need to generate the contract address in order to fetch the bytecode
    if to is None:
        baddr = mk_contract_address(tx['from'], tx['nonce'])
        to = '0x%s' % encode_hex(baddr).decode()

    cache = {}
    for addr in getAddresses(ops, to):
        if addr in cache:
            c = cache[addr]

            if create and addr == to:
                if addr + '_created' in c:
                    c = cache[addr + '_created']
                else:
                    # if it's cached, then the contract is already created so we need
                    # to return the Contract instance w/ create = False
                    newc = object.__new__(Contract)
                    newc.__dict__ = c.__dict__.copy()
                    newc.create = False
                    c = newc
                    cache[addr + '_created'] = c

        else:
            acc = api.getAccountInfo(addr, blnum)
            c = findContractForBytecode(contracts, acc['code'])
            cache[addr] = c
            if not c:
                print("Couldn't find contract for address {}".format(addr))
            if c and create and addr == to:
                c.create = True

        contract_stack.append(Context(addr, c))

    # contractTexts are not loaded by default, the following will
    # load the contractTexts and populate the sourceCache for the contract
    # corresponding to each op in this tx, greatly improving the response
    # time when quickly moving through the opviewer ui
    i = 0
    while i < len(ops):
        pc = ops[i]['pc']
        contract_stack[i].getSourceCode(pc)
        i += 1

    return contract_stack


def getAddresses(ops, original_contract):
    """ determine the address of the sourceCode for each operation
     Returns an array of addresses, 1 for each op in ops
     """
    addresses = []

    cur_depth = None
    prev_depth = None
    prev_op = None
    cur_address = original_contract
    place_holders = []  # stores the index of where CREATE op addr should go in the addr_stack
    depth_to_addr = {}  # mapping to track depth to an addr so we can easily push the current addr on the stack when we return from a call

    step = 0
    for o in ops:
        if not 'depth' in o.keys():
            # We're done here
            break

        # Address tracking
        cur_depth = o['depth']

        # depth may not always start at 1. ganache-cli starts at 0
        # this will set the prev_depth from the first op
        if not prev_op:
            prev_depth = cur_depth
            depth_to_addr[cur_depth] = cur_address

        if cur_depth > prev_depth:
            # Made it into a call-variant
            # All call-lookalikes are 'gas,address,value' on stack,
            # so address is second item of prev line
            #
            # There's one exception, though; CREATE
            # With a CREATE, we don't know the address until after the RETURN
            # so we push a placeholder and update on the Return
            if prev_op['op'] == 0xf0:
                cur_address = None
                place_holders.append([len(addresses)])
            else:
                cur_address = prev_op['stack'][-2]

            depth_to_addr[cur_depth] = cur_address

        if cur_depth < prev_depth:
            # RETURN op. we now know the prev_depth address, so add to context
            if cur_address is None and prev_op['op'] == 0xf3:
                prev_address = o['stack'][-1]
                for i in place_holders.pop():
                    addresses[i] = prev_address
            # Returned from a call
            cur_address = depth_to_addr[cur_depth]

        if not cur_address:
            place_holders[-1].append(len(addresses))

        addresses.append(cur_address)
        prev_op = o
        prev_depth = cur_depth

    # handle debug_traceTransaction output
    def fixAddr(a):
        if a and len(a) > 40:
            if (a.startswith('0x') and len(a) == 42):
                return a
            else:
                return "0x%s" % a[24:]

    addresses = [fixAddr(a) for a in addresses]

    return addresses


def findContractForBytecode(contracts, bytecode):
    if bytecode.startswith('0x'):
        bytecode = bytecode[2:]

    for c in contracts:
        # ignore last 34 bytes which is just metadata
        if c.bin and c.bin[:68] == bytecode[:68] or c.binRuntime and c.binRuntime[:68] == bytecode[:68]:
            return c

    return None


class Context(object):
    def __init__(self, address, contract):
        self.address = address
        self.contract = contract

    def getSourceCode(self, pc):
        if self.contract:
            return self.contract.getSourceCode(pc)

        return "Missing Contract", (0, 0)

    @property
    def name(self):
        return self.contract.name if self.contract else "?"
